import abc
import asyncio
import typing
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    no_type_check,
)

from mode import Seconds, ServiceT
from mode.utils.trees import NodeT

from .channels import ChannelT
from .core import K
from .events import EventT
from .models import FieldDescriptorT, ModelArg
from .topics import TopicT
from .tuples import TP

if typing.TYPE_CHECKING:
    from .app import AppT as _AppT
    from .join import JoinT as _JoinT
    from .serializers import SchemaT as _SchemaT
else:
    class _AppT: ...     # noqa
    class _JoinT: ...    # noqa
    class _SchemaT: ...  # noqa

__all__ = [
    'Processor',
    'GroupByKeyArg',
    'StreamT',
    'T',
    'T_co',
    'T_contra',
]

# Used for typing StreamT[Withdrawal]
T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)
T_contra = TypeVar('T_contra', contravariant=True)

Processor = Callable[[T], Union[T, Awaitable[T]]]

#: Type of the `key` argument to `Stream.group_by()`
GroupByKeyArg = Union[FieldDescriptorT, Callable[[T], K]]


class JoinableT(abc.ABC):

    @abc.abstractmethod
    def combine(self, *nodes: 'JoinableT', **kwargs: Any) -> 'StreamT':
        ...

    @abc.abstractmethod
    def join(self, *fields: FieldDescriptorT) -> 'StreamT':
        ...

    @abc.abstractmethod
    def left_join(self, *fields: FieldDescriptorT) -> 'StreamT':
        ...

    @abc.abstractmethod
    def inner_join(self, *fields: FieldDescriptorT) -> 'StreamT':
        ...

    @abc.abstractmethod
    def outer_join(self, *fields: FieldDescriptorT) -> 'StreamT':
        ...

    @abc.abstractmethod
    def __and__(self, other: Any) -> Any:
        ...

    @abc.abstractmethod
    def contribute_to_stream(self, active: 'StreamT') -> None:
        ...

    @abc.abstractmethod
    async def remove_from_stream(self, stream: 'StreamT') -> None:
        ...

    @abc.abstractmethod
    def _human_channel(self) -> str:
        ...


class StreamT(AsyncIterable[T_co], JoinableT, ServiceT):

    app: _AppT
    channel: AsyncIterator[T_co]
    outbox: Optional[asyncio.Queue] = None
    join_strategy: Optional[_JoinT] = None
    task_owner: Optional[asyncio.Task] = None
    current_event: Optional[EventT] = None
    active_partitions: Optional[Set[TP]] = None
    concurrency_index: Optional[int] = None
    enable_acks: bool = True
    prefix: str = ''

    # List of combined streams/tables after ret = (s1 & s2) combined them.
    # AFter this ret.combined == [s1, s2]
    combined: List[JoinableT]

    # group_by, through, etc. sets this, and it means the
    # active stream (the one the agent would be reading from) can be found
    # by walking the path of links::
    #    >>> node = stream
    #    >>> while node._next:
    #    ...     node = node._next
    # which is also what .get_active_stream() gives
    _next: Optional['StreamT'] = None
    _prev: Optional['StreamT'] = None

    @abc.abstractmethod
    def __init__(self,
                 channel: AsyncIterator[T_co] = None,
                 *,
                 app: _AppT = None,
                 processors: Iterable[Processor[T]] = None,
                 combined: List[JoinableT] = None,
                 on_start: Callable = None,
                 join_strategy: _JoinT = None,
                 beacon: NodeT = None,
                 concurrency_index: int = None,
                 prev: 'StreamT' = None,
                 active_partitions: Set[TP] = None,
                 enable_acks: bool = True,
                 prefix: str = '',
                 loop: asyncio.AbstractEventLoop = None) -> None:
        ...

    @abc.abstractmethod
    def get_active_stream(self) -> 'StreamT':
        ...

    @abc.abstractmethod
    def add_processor(self, processor: Processor[T]) -> None:
        ...

    @abc.abstractmethod
    def info(self) -> Mapping[str, Any]:
        ...

    @abc.abstractmethod
    def clone(self, **kwargs: Any) -> 'StreamT':
        ...

    @abc.abstractmethod
    @no_type_check
    async def items(self) -> AsyncIterator[Tuple[K, T_co]]:
        ...

    @abc.abstractmethod
    @no_type_check
    async def events(self) -> AsyncIterable[EventT]:
        ...

    @abc.abstractmethod
    @no_type_check
    async def take(self, max_: int,
                   within: Seconds) -> AsyncIterable[Sequence[T_co]]:
        ...

    @abc.abstractmethod
    def enumerate(self, start: int = 0) -> AsyncIterable[Tuple[int, T_co]]:
        ...

    @abc.abstractmethod
    def through(self, channel: Union[str, ChannelT]) -> 'StreamT':
        ...

    @abc.abstractmethod
    def echo(self, *channels: Union[str, ChannelT]) -> 'StreamT':
        ...

    @abc.abstractmethod
    def group_by(self,
                 key: GroupByKeyArg,
                 *,
                 name: str = None,
                 topic: TopicT = None) -> 'StreamT':
        ...

    @abc.abstractmethod
    def derive_topic(self,
                     name: str,
                     *,
                     schema: _SchemaT = None,
                     key_type: ModelArg = None,
                     value_type: ModelArg = None,
                     prefix: str = '',
                     suffix: str = '') -> TopicT:
        ...

    @abc.abstractmethod
    async def throw(self, exc: BaseException) -> None:
        ...

    @abc.abstractmethod
    def __copy__(self) -> 'StreamT':
        ...

    @abc.abstractmethod
    def __iter__(self) -> Any:
        ...

    @abc.abstractmethod
    def __next__(self) -> T:
        ...

    @abc.abstractmethod
    def __aiter__(self) -> AsyncIterator[T_co]:
        ...

    @abc.abstractmethod
    async def ack(self, event: EventT) -> bool:
        ...



import abc
import typing
from typing import Any, Callable, Generic, Optional, Tuple, TypeVar

from .codecs import CodecArg
from .core import K, OpenHeadersArg, V

if typing.TYPE_CHECKING:
    from .app import AppT as _AppT
    from .models import ModelArg as _ModelArg
    from .tuples import Message as _Message
else:
    class _AppT: ...      # noqa
    class _ModelArg: ...  # noqa
    class _Message: ...   # noqa

__all__ = ['RegistryT', 'SchemaT']

KT = TypeVar('KT')
VT = TypeVar('VT')


class RegistryT(abc.ABC):

    key_serializer: CodecArg
    value_serializer: CodecArg

    @abc.abstractmethod
    def __init__(self,
                 key_serializer: CodecArg = None,
                 value_serializer: CodecArg = 'json') -> None:
        ...

    @abc.abstractmethod
    def loads_key(self,
                  typ: Optional[_ModelArg],
                  key: Optional[bytes],
                  *,
                  serializer: CodecArg = None) -> K:
        ...

    @abc.abstractmethod
    def loads_value(self,
                    typ: Optional[_ModelArg],
                    value: Optional[bytes],
                    *,
                    serializer: CodecArg = None) -> Any:
        ...

    @abc.abstractmethod
    def dumps_key(self,
                  typ: Optional[_ModelArg],
                  key: K,
                  *,
                  serializer: CodecArg = None) -> Optional[bytes]:
        ...

    @abc.abstractmethod
    def dumps_value(self,
                    typ: Optional[_ModelArg],
                    value: V,
                    *,
                    serializer: CodecArg = None) -> Optional[bytes]:
        ...


class SchemaT(Generic[KT, VT]):

    key_type: Optional[_ModelArg] = None
    value_type: Optional[_ModelArg] = None

    key_serializer: Optional[CodecArg] = None
    value_serializer: Optional[CodecArg] = None

    allow_empty: bool = False

    def __init__(self, *,
                 key_type: _ModelArg = None,
                 value_type: _ModelArg = None,
                 key_serializer: CodecArg = None,
                 value_serializer: CodecArg = None,
                 allow_empty: bool = None) -> None:
        ...

    @abc.abstractmethod
    def update(self, *,
               key_type: _ModelArg = None,
               value_type: _ModelArg = None,
               key_serializer: CodecArg = None,
               value_serializer: CodecArg = None,
               allow_empty: bool = None) -> None:
        ...

    @abc.abstractmethod
    def loads_key(self, app: _AppT, message: _Message, *,
                  loads: Callable = None,
                  serializer: CodecArg = None) -> KT:
        ...

    @abc.abstractmethod
    def loads_value(self, app: _AppT, message: _Message, *,
                    loads: Callable = None,
                    serializer: CodecArg = None) -> VT:
        ...

    @abc.abstractmethod
    def dumps_key(self, app: _AppT, key: K, *,
                  serializer: CodecArg = None,
                  headers: OpenHeadersArg) -> Tuple[Any, OpenHeadersArg]:
        ...

    @abc.abstractmethod
    def dumps_value(self, app: _AppT, value: V, *,
                    serializer: CodecArg = None,
                    headers: OpenHeadersArg) -> Tuple[Any, OpenHeadersArg]:
        ...

    @abc.abstractmethod
    def on_dumps_key_prepare_headers(
            self, key: V, headers: OpenHeadersArg) -> OpenHeadersArg:
        ...

    @abc.abstractmethod
    def on_dumps_value_prepare_headers(
            self, value: V, headers: OpenHeadersArg) -> OpenHeadersArg:
        ...



import os
import faust

# make sure the event loop is used as early as possible.
os.environ.setdefault('FAUST_LOOP', 'eventlet')

# set the default Django settings module for the 'faust' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'proj.settings')

app = faust.App('django-proj', autodiscover=True, origin='faustapp')


@app.on_configured.connect
def configure_from_settings(app, conf, **kwargs):
    from django.conf import settings
    conf.broker = settings.FAUST_BROKER_URL
    conf.store = settings.FAUST_STORE_URL


def main():
    app.main()


if __name__ == '__main__':
    main()



"""HTTP and related errors."""
import http
from typing import Any, Dict, cast
from faust.exceptions import FaustError

__all__ = [
    'WebError',
    'ServerError',
    'ValidationError',
    'ParseError',
    'AuthenticationFailed',
    'NotAuthenticated',
    'PermissionDenied',
    'NotFound',
    'MethodNotAllowed',
    'NotAcceptable',
    'UnsupportedMediaType',
    'Throttled',
]


class WebError(FaustError):
    """Web related error.

    Web related errors will have a status :attr:`code`,
    and a :attr:`detail` for the human readable error string.

    It may also keep :attr:`extra_context`.
    """

    code: int = cast(int, None)
    detail: str = 'Default not set on class'
    extra_context: Dict

    def __init__(self, detail: str = None, *,
                 code: int = None,
                 **extra_context: Any) -> None:
        if detail:
            self.detail = detail
        if code:
            self.code = code
        self.extra_context = extra_context
        super().__init__(self, detail, code, extra_context)


class ServerError(WebError):
    """Internal Server Error (500)."""

    code = http.HTTPStatus.INTERNAL_SERVER_ERROR
    detail = 'Internal server error.'


class ValidationError(WebError):
    """Invalid input in POST data (400)."""

    code = http.HTTPStatus.BAD_REQUEST
    detail = 'Invalid input.'


class ParseError(WebError):
    """Malformed request (400)."""

    code = http.HTTPStatus.BAD_REQUEST
    detail = 'Malformed request.'


class AuthenticationFailed(WebError):
    """Incorrect authentication credentials (401)."""

    code = http.HTTPStatus.UNAUTHORIZED
    detail = 'Incorrect authentication credentials'


class NotAuthenticated(WebError):
    """Authentication credentials were not provided (401)."""

    code = http.HTTPStatus.UNAUTHORIZED
    detail = 'Authentication credentials were not provided.'


class PermissionDenied(WebError):
    """No permission to perform action (403)."""

    code = http.HTTPStatus.FORBIDDEN
    detail = 'You do not have permission to perform this action.'


class NotFound(WebError):
    """Resource not found (404)."""

    code = http.HTTPStatus.NOT_FOUND
    detail = 'Not found.'


class MethodNotAllowed(WebError):
    """HTTP Method not allowed (405)."""

    code = http.HTTPStatus.METHOD_NOT_ALLOWED
    detail = 'Method not allowed.'


class NotAcceptable(WebError):
    """Not able to satisfy the request ``Accept`` header (406)."""

    code = http.HTTPStatus.NOT_ACCEPTABLE
    detail = 'Could not satisfy the request Accept header.'


class UnsupportedMediaType(WebError):
    """Request contains unsupported media type (415)."""

    code = http.HTTPStatus.UNSUPPORTED_MEDIA_TYPE
    detail = 'Unsupported media type in request.'


class Throttled(WebError):
    """Client is sending too many requests to server (429)."""

    code = http.HTTPStatus.TOO_MANY_REQUESTS
    detail = 'Request was throttled.'


class ServiceUnavailable(WebError):
    """Service is temporarily unavailable (503)."""

    code = http.HTTPStatus.SERVICE_UNAVAILABLE
    detail = 'Service unavailable. Try again later.'



import pytest
from collections import Counter
from pprint import pformat
from typing import List, NamedTuple
from faust.exceptions import ImproperlyConfigured
from faust.types import EventT, StreamT
from .helpers import AgentCase


class _param(NamedTuple):
    partitions: List[int]
    num_messages: int
    isolated_partitions: bool = False


@pytest.mark.asyncio
@pytest.mark.parametrize(_param._fields, [
    _param([0], 100, isolated_partitions=True),
    _param([0], 100, isolated_partitions=False),
    _param(list(range(10)), 100, isolated_partitions=True),
    _param(list(range(10)), 100, isolated_partitions=False),
    _param(list(range(100)), 1000, isolated_partitions=True),
    _param(list(range(10)), 1000, isolated_partitions=True),
    _param(list(range(10)), 1000, isolated_partitions=False),
])
async def test_agent_isolated_partitions(
        partitions,
        num_messages,
        isolated_partitions,
        *,
        app,
        logging):
    await AgentIsolatedCase.run_test(
        app=app,
        num_messages=num_messages,
        concurrency=1,
        partitions=partitions,
        isolated_partitions=isolated_partitions,
    )


@pytest.mark.asyncio
async def test_agent_isolated_partitions__concurrency(*, app, logging):
    with pytest.raises(ImproperlyConfigured):
        await AgentIsolatedCase.run_test(
            app=app,
            concurrency=2,
            isolated_partitions=True,
        )


class AgentIsolatedCase(AgentCase):
    name = 'test_agent_isolated_partitions'

    def on_init(self) -> None:
        # isolated_partitions=True requires concurrency=1
        self.processed_by_tp = Counter()

    async def on_agent_event(
            self, stream: StreamT, event: EventT) -> None:
        # when an event is processed by our agent
        if self.isolated_partitions:
            # make sure the event tp matches the tps for this actor
            assert stream.active_partitions is not None
            assert len(stream.active_partitions) == 1
            assert event.message.tp in stream.active_partitions
        else:
            assert stream.active_partitions is None

        # tests: ordering/uniform distribution
        assert event.message.tp == next(self.expected_tp)

        self.processed_by_tp[event.message.tp] += 1

    async def assert_success(self) -> None:
        await super().assert_success()
        self.log.info(
            '- Final TopicPartition distribution by coroutine:\n%s',
            pformat([
                (tp.partition, total)
                for tp, total in self.processed_by_tp.most_common()
            ]),
        )
        max_ = None
        for _, total in self.processed_by_tp.most_common():
            if max_ is None:
                max_ = total
            assert total == max_



"""Message transport using :pypi:`confluent_kafka`."""
import asyncio
import typing

from collections import defaultdict
from typing import (
    Any,
    Awaitable,
    Callable,
    ClassVar,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Set,
    Type,
    cast,
)

from mode import Service, get_logger
from mode.threads import QueueServiceThread
from mode.utils.futures import notify
from mode.utils.times import Seconds, want_seconds
from yarl import URL

from faust.exceptions import ConsumerNotStarted, ProducerSendError
from faust.transport import base
from faust.transport.consumer import (
    ConsumerThread,
    RecordMap,
    ThreadDelegateConsumer,
    ensure_TP,
    ensure_TPset,
)
from faust.types import AppT, ConsumerMessage, HeadersArg, RecordMetadata, TP
from faust.types.transports import ConsumerT, ProducerT

import confluent_kafka
from confluent_kafka import TopicPartition as _TopicPartition
from confluent_kafka import KafkaException

if typing.TYPE_CHECKING:
    from confluent_kafka import Consumer as _Consumer
    from confluent_kafka import Producer as _Producer
    from confluent_kafka import Message as _Message
else:
    class _Consumer: ...  # noqa
    class _Producer: ...  # noqa
    class _Message: ...   # noqa

__all__ = ['Consumer', 'Producer', 'Transport']


logger = get_logger(__name__)


def server_list(urls: List[URL], default_port: int) -> str:
    default_host = '127.0.0.1'
    return ','.join([
        f'{u.host or default_host}:{u.port or default_port}' for u in urls])


class Consumer(ThreadDelegateConsumer):
    """Kafka consumer using :pypi:`confluent_kafka`."""

    logger = logger

    def _new_consumer_thread(self) -> ConsumerThread:
        return ConfluentConsumerThread(
            self, loop=self.loop, beacon=self.beacon)

    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 30.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        """Create topic on broker."""
        return  # XXX
        await self._thread.create_topic(
            topic,
            partitions,
            replication,
            config=config,
            timeout=int(want_seconds(timeout) * 1000.0),
            retention=int(want_seconds(retention) * 1000.0),
            compacting=compacting,
            deleting=deleting,
            ensure_created=ensure_created,
        )

    def _to_message(self, tp: TP, record: Any) -> ConsumerMessage:
        # convert timestamp to seconds from int milliseconds.
        timestamp_type: int
        timestamp: Optional[int]
        timestamp_type, timestamp = record.timestamp()
        timestamp_s: float = cast(float, None)
        if timestamp is not None:
            timestamp_s = timestamp / 1000.0
        key = record.key()
        key_size = len(key) if key is not None else 0
        value = record.value()
        value_size = len(value) if value is not None else 0
        return ConsumerMessage(
            record.topic(),
            record.partition(),
            record.offset(),
            timestamp_s,
            timestamp_type,
            [],  # headers
            key,
            value,
            None,
            key_size,
            value_size,
            tp,
        )

    def _new_topicpartition(self, topic: str, partition: int) -> TP:
        return cast(TP, _TopicPartition(topic, partition))


class ConfluentConsumerThread(ConsumerThread):
    """Thread managing underlying :pypi:`confluent_kafka` consumer."""

    _consumer: Optional[_Consumer] = None
    _assigned: bool = False

    async def on_start(self) -> None:
        self._consumer = self._create_consumer(loop=self.thread_loop)

    def _create_consumer(
            self,
            loop: asyncio.AbstractEventLoop) -> _Consumer:
        transport = cast(Transport, self.transport)
        if self.app.client_only:
            return self._create_client_consumer(transport, loop=loop)
        else:
            return self._create_worker_consumer(transport, loop=loop)

    def _create_worker_consumer(
            self,
            transport: 'Transport',
            loop: asyncio.AbstractEventLoop) -> _Consumer:
        conf = self.app.conf
        self._assignor = self.app.assignor

        # XXX parition.assignment.strategy is string
        # need to write C wrapper for this
        # 'partition.assignment.strategy': [self._assignor]
        return confluent_kafka.Consumer({
            'bootstrap.servers': server_list(
                transport.url, transport.default_port),
            'group.id': conf.id,
            'client.id': conf.broker_client_id,
            'default.topic.config': {
                'auto.offset.reset': 'earliest',
            },
            'enable.auto.commit': False,
            'fetch.max.bytes': conf.consumer_max_fetch_size,
            'request.timeout.ms': int(conf.broker_request_timeout * 1000.0),
            'check.crcs': conf.broker_check_crcs,
            'session.timeout.ms': int(conf.broker_session_timeout * 1000.0),
            'heartbeat.interval.ms': int(
                conf.broker_heartbeat_interval * 1000.0),
        })

    def _create_client_consumer(
            self,
            transport: 'Transport',
            loop: asyncio.AbstractEventLoop) -> _Consumer:
        conf = self.app.conf
        return confluent_kafka.Consumer({
            'bootstrap.servers': server_list(
                transport.url, transport.default_port),
            'client.id': conf.broker_client_id,
            'enable.auto.commit': True,
            'default.topic.config': {
                'auto.offset.reset': 'earliest',
            },
        })

    def close(self) -> None:
        ...

    async def subscribe(self, topics: Iterable[str]) -> None:
        # XXX pattern does not work :/
        await self.call_thread(
            self._ensure_consumer().subscribe,
            topics=list(topics),
            on_assign=self._on_assign,
            on_revoke=self._on_revoke,
        )
        while not self._assigned:
            self.log.info('Still waiting for assignment...')
            self._ensure_consumer().poll(timeout=1)

    def _on_assign(self,
                   consumer: _Consumer,
                   assigned: List[_TopicPartition]) -> None:
        self._assigned = True
        self.thread_loop.run_until_complete(
            self.on_partitions_assigned(
                {TP(tp.topic, tp.partition) for tp in assigned}))

    def _on_revoke(self,
                   consumer: _Consumer,
                   revoked: List[_TopicPartition]) -> None:
        self.thread_loop.run_until_complete(
            self.on_partitions_revoked(
                {TP(tp.topic, tp.partition) for tp in revoked}))

    async def seek_to_committed(self) -> Mapping[TP, int]:
        return await self.call_thread(self._seek_to_committed)

    async def _seek_to_committed(self) -> Mapping[TP, int]:
        consumer = self._ensure_consumer()
        assignment = consumer.assignment()
        committed = consumer.committed(assignment)
        for tp in committed:
            consumer.seek(tp)
        return {ensure_TP(tp): tp.offset for tp in committed}

    async def _committed_offsets(
            self, partitions: List[TP]) -> MutableMapping[TP, int]:
        consumer = self._ensure_consumer()
        committed = consumer.committed(
            [_TopicPartition(tp[0], tp[1]) for tp in partitions])
        return {
            TP(tp.topic, tp.partition): tp.offset
            for tp in committed
        }

    async def commit(self, tps: Mapping[TP, int]) -> bool:
        self.call_thread(
            self._ensure_consumer().commit,
            offsets=[
                _TopicPartition(tp.topic, tp.partition, offset=offset)
                for tp, offset in tps.items()
            ],
            asynchronous=False,
        )
        return True

    async def position(self, tp: TP) -> Optional[int]:
        return await self.call_thread(
            self._ensure_consumer().position, tp)

    async def seek_to_beginning(self, *partitions: _TopicPartition) -> None:
        await self.call_thread(
            self._ensure_consumer().seek_to_beginning, *partitions)

    async def seek_wait(self, partitions: Mapping[TP, int]) -> None:
        consumer = self._ensure_consumer()
        await self.call_thread(self._seek_wait, consumer, partitions)

    async def _seek_wait(self,
                         consumer: Consumer,
                         partitions: Mapping[TP, int]) -> None:
        for tp, offset in partitions.items():
            self.log.dev('SEEK %r -> %r', tp, offset)
            consumer.seek(tp, offset)
        await asyncio.gather(*[
            consumer.position(tp) for tp in partitions
        ])

    def seek(self, partition: TP, offset: int) -> None:
        self._ensure_consumer().seek(partition, offset)

    def assignment(self) -> Set[TP]:
        return ensure_TPset(self._ensure_consumer().assignment())

    def highwater(self, tp: TP) -> int:
        _, hw = self._ensure_consumer().get_watermark_offsets(
            _TopicPartition(tp.topic, tp.partition), cached=True)
        return hw

    def topic_partitions(self, topic: str) -> Optional[int]:
        # XXX NotImplemented
        return None

    async def earliest_offsets(self,
                               *partitions: TP) -> MutableMapping[TP, int]:
        if not partitions:
            return {}
        return await self.call_thread(self._earliest_offsets, partitions)

    async def _earliest_offsets(
            self, partitions: List[TP]) -> MutableMapping[TP, int]:
        consumer = self._ensure_consumer()
        return {
            tp: consumer.get_watermark_offsets(
                _TopicPartition(tp[0], tp[1]))[0]
            for tp in partitions
        }

    async def highwaters(self, *partitions: TP) -> MutableMapping[TP, int]:
        if not partitions:
            return {}
        return await self.call_thread(self._highwaters, partitions)

    async def _highwaters(
            self, partitions: List[TP]) -> MutableMapping[TP, int]:
        consumer = self._ensure_consumer()
        return {
            tp: consumer.get_watermark_offsets(
                _TopicPartition(tp[0], tp[1]))[1]
            for tp in partitions
        }

    def _ensure_consumer(self) -> _Consumer:
        if self._consumer is None:
            raise ConsumerNotStarted('Consumer thread not yet started')
        return self._consumer

    async def getmany(self,
                      active_partitions: Optional[Set[TP]],
                      timeout: float) -> RecordMap:
        # Implementation for the Fetcher service.
        _consumer = self._ensure_consumer()
        messages = await self.call_thread(
            _consumer.consume,
            num_messages=10000,
            timeout=timeout,
        )
        records: RecordMap = defaultdict(list)
        for message in messages:
            tp = TP(message.topic(), message.partition())
            records[tp].append(message)
        return records

    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 30.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        return  # XXX

    def key_partition(self,
                      topic: str,
                      key: Optional[bytes],
                      partition: int = None) -> Optional[int]:
        raise NotImplementedError('TODO')  # TODO XXX


class ProducerProduceFuture(asyncio.Future):

    def set_from_on_delivery(self,
                             err: Optional[BaseException],
                             msg: _Message) -> None:
        if err:
            # XXX Not sure what err' is here, hopefully it's an exception
            # object and not a string [ask].
            self.set_exception(err)
        else:
            metadata: RecordMetadata = self.message_to_metadata(msg)
            self.set_result(metadata)

    def message_to_metadata(self, message: _Message) -> RecordMetadata:
        topic, partition = tp = TP(message.topic(), message.partition())
        return RecordMetadata(topic, partition, tp, message.offset())


class ProducerThread(QueueServiceThread):
    """Thread managing underlying :pypi:`confluent_kafka` producer."""

    app: AppT
    producer: 'Producer'
    transport: 'Transport'
    _producer: Optional[_Producer] = None
    _flush_soon: Optional[asyncio.Future] = None

    def __init__(self, producer: 'Producer', **kwargs: Any) -> None:
        self.producer = producer
        self.transport = cast(Transport, self.producer.transport)
        self.app = self.transport.app
        super().__init__(**kwargs)

    async def on_start(self) -> None:
        self._producer = confluent_kafka.Producer({
            'bootstrap.servers': server_list(
                self.transport.url, self.transport.default_port),
            'client.id': self.app.conf.broker_client_id,
            'max.in.flight.requests.per.connection': 1,
        })

    async def flush(self) -> None:
        if self._producer is not None:
            self._producer.flush()

    async def on_thread_stop(self) -> None:
        if self._producer is not None:
            self._producer.flush()

    def produce(self, topic: str, key: bytes, value: bytes, partition: int,
                on_delivery: Callable) -> None:
        if self._producer is None:
            raise RuntimeError('Producer not started')
        if partition is not None:
            self._producer.produce(
                topic, key, value, partition, on_delivery=on_delivery,
            )
        else:
            self._producer.produce(
                topic, key, value, on_delivery=on_delivery,
            )
        notify(self._flush_soon)

    @Service.task
    async def _background_flush(self) -> None:
        producer = cast(_Producer, self._producer)
        _size = producer.__len__
        _flush = producer.flush
        _poll = producer.poll
        _sleep = self.sleep
        _create_future = self.loop.create_future
        while not self.should_stop:
            if not _size():
                flush_soon = self._flush_soon
                if flush_soon is None:
                    flush_soon = self._flush_soon = _create_future()
                stopped: bool = False
                try:
                    stopped = await self.wait_for_stopped(
                        flush_soon, timeout=1.0)
                finally:
                    self._flush_soon = None
                if not stopped:
                    _flush(timeout=100)
                    _poll(timeout=1)
                    await _sleep(0)


class Producer(base.Producer):
    """Kafka producer using :pypi:`confluent_kafka`."""

    logger = logger

    _producer_thread: ProducerThread
    _quick_produce: Any = None

    def __post_init__(self) -> None:
        self._producer_thread = ProducerThread(
            self, loop=self.loop, beacon=self.beacon)
        self._quick_produce = self._producer_thread.produce

    async def on_restart(self) -> None:
        """Call when producer is restarting."""
        self.on_init()

    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 20.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        """Create topic on broker."""
        return  # XXX
        _retention = (int(want_seconds(retention) * 1000.0)
                      if retention else None)
        await cast(Transport, self.transport)._create_topic(
            self,
            self._producer.client,
            topic,
            partitions,
            replication,
            config=config,
            timeout=int(want_seconds(timeout) * 1000.0),
            retention=_retention,
            compacting=compacting,
            deleting=deleting,
            ensure_created=ensure_created,
        )

    async def on_start(self) -> None:
        """Call when producer is starting."""
        await self._producer_thread.start()
        await self.sleep(0.5)  # cannot remember why, necessary? [ask]

    async def on_stop(self) -> None:
        """Call when producer is stopping."""
        await self._producer_thread.stop()

    async def send(self, topic: str, key: Optional[bytes],
                   value: Optional[bytes],
                   partition: Optional[int],
                   timestamp: Optional[float],
                   headers: Optional[HeadersArg],
                   *,
                   transactional_id: str = None) -> Awaitable[RecordMetadata]:
        """Send message for future delivery."""
        fut = ProducerProduceFuture(loop=self.loop)
        self._quick_produce(
            topic, value, key, partition,
            timestamp=int(timestamp * 1000) if timestamp else timestamp,
            on_delivery=fut.set_from_on_delivery,
        )
        return cast(Awaitable[RecordMetadata], fut)
        try:
            return cast(Awaitable[RecordMetadata], await self._producer.send(
                topic, value, key=key, partition=partition))
        except KafkaException as exc:
            raise ProducerSendError(f'Error while sending: {exc!r}') from exc

    async def send_and_wait(self, topic: str, key: Optional[bytes],
                            value: Optional[bytes],
                            partition: Optional[int],
                            timestamp: Optional[float],
                            headers: Optional[HeadersArg],
                            *,
                            transactional_id: str = None) -> RecordMetadata:
        """Send message and wait for it to be delivered to broker(s)."""
        fut = await self.send(
            topic, key, value, partition, timestamp, headers,
        )
        return await fut

    async def flush(self) -> None:
        """Flush producer buffer.

        This will wait until the producer has written
        all buffered up messages to any connected brokers.
        """
        await self._producer_thread.flush()

    def key_partition(self, topic: str, key: bytes) -> TP:
        """Return topic and partition destination for key."""
        raise NotImplementedError()


class Transport(base.Transport):
    """Kafka transport using :pypi:`confluent_kafka`."""

    Consumer: ClassVar[Type[ConsumerT]] = Consumer
    Producer: ClassVar[Type[ProducerT]] = Producer

    default_port = 9092
    driver_version = f'confluent_kafka={confluent_kafka.__version__}'

    def _topic_config(self,
                      retention: int = None,
                      compacting: bool = None,
                      deleting: bool = None) -> MutableMapping[str, Any]:
        config: MutableMapping[str, Any] = {}
        cleanup_flags: Set[str] = set()
        if compacting:
            cleanup_flags |= {'compact'}
        if deleting:
            cleanup_flags |= {'delete'}
        if cleanup_flags:
            config['cleanup.policy'] = ','.join(sorted(cleanup_flags))
        if retention:
            config['retention.ms'] = retention
        return config



import abc
import typing
from typing import Any, Awaitable, Generic, Mapping, Optional, TypeVar, Union
from mode.utils.typing import AsyncContextManager
from .codecs import CodecArg
from .core import HeadersArg, K, V
from .tuples import Message, MessageSentCallback, RecordMetadata

if typing.TYPE_CHECKING:
    from .app import AppT as _AppT
    from .channels import ChannelT as _ChannelT
    from .serializers import SchemaT as _SchemaT
else:
    class _AppT: ...  # noqa
    class _ChannelT: ...  # noqa
    class _SchemaT: ...   # noqa

T = TypeVar('T')


class EventT(Generic[T], AsyncContextManager):

    app: _AppT
    key: K
    value: V
    headers: Mapping
    message: Message
    acked: bool

    __slots__ = ('app', 'key', 'value', 'headers', 'message', 'acked')

    @abc.abstractmethod
    def __init__(self,
                 app: _AppT,
                 key: K,
                 value: V,
                 headers: Optional[HeadersArg],
                 message: Message) -> None:
        ...

    @abc.abstractmethod
    async def send(self,
                   channel: Union[str, _ChannelT],
                   key: K = None,
                   value: V = None,
                   partition: int = None,
                   timestamp: float = None,
                   headers: HeadersArg = None,
                   schema: _SchemaT = None,
                   key_serializer: CodecArg = None,
                   value_serializer: CodecArg = None,
                   callback: MessageSentCallback = None,
                   force: bool = False) -> Awaitable[RecordMetadata]:
        ...

    @abc.abstractmethod
    async def forward(self,
                      channel: Union[str, _ChannelT],
                      key: Any = None,
                      value: Any = None,
                      partition: int = None,
                      timestamp: float = None,
                      headers: HeadersArg = None,
                      schema: _SchemaT = None,
                      key_serializer: CodecArg = None,
                      value_serializer: CodecArg = None,
                      callback: MessageSentCallback = None,
                      force: bool = False) -> Awaitable[RecordMetadata]:
        ...

    @abc.abstractmethod
    def ack(self) -> bool:
        ...



import pytest
from faust import Event, Record, Stream
from faust.joins import InnerJoin, Join, LeftJoin, OuterJoin, RightJoin
from mode.utils.mocks import Mock


class User(Record):
    id: str
    name: str


@pytest.mark.asyncio
@pytest.mark.parametrize('join_cls,fields', [
    (Join, (User.id, User.name)),
    (InnerJoin, (User.id, User.name)),
    (LeftJoin, (User.id, User.name)),
    (OuterJoin, (User.id, User.name)),
    (RightJoin, (User.id, User.name)),
])
async def test_Join(join_cls, fields):
    stream = Mock(name='stream', autospec=Stream)
    j = join_cls(stream=stream, fields=fields)
    assert j.fields
    assert j.stream is stream

    with pytest.raises(NotImplementedError):
        await j.process(Mock(name='event', autospec=Event))


def test_eq_ne():
    stream = Mock(name='stream', autospec=Stream)
    assert (InnerJoin(stream=stream, fields=(User.id,)) ==
            InnerJoin(stream=stream, fields=(User.id,)))
    assert (InnerJoin(stream=stream, fields=(User.id,)) !=
            Join(stream=stream, fields=(User.id,)))
    assert (InnerJoin(stream=stream, fields=(User.name,)) !=
            InnerJoin(stream=stream, fields=(User.id,)))



import abc
import asyncio
import typing
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Generic,
    Optional,
    Set,
    TypeVar,
)

from mode import Seconds
from mode.utils.futures import stampede
from mode.utils.queues import ThrowableQueue

from .codecs import CodecArg
from .core import HeadersArg, K, V
from .tuples import (
    FutureMessage,
    Message,
    MessageSentCallback,
    RecordMetadata,
    TP,
)

_T = TypeVar('_T')
_T_contra = TypeVar('_T_contra', contravariant=True)

if typing.TYPE_CHECKING:
    from .app import AppT as _AppT
    from .events import EventT as _EventT
    from .models import ModelArg as _ModelArg
    from .serializers import SchemaT as _SchemaT
    from .streams import StreamT as _StreamT
else:
    class _AppT: ...                          # noqa
    class _EventT(Generic[_T]): ...           # noqa
    class _ModelArg: ...                      # noqa
    class _SchemaT: ...                       # noqa
    class _StreamT: ...                       # noqa


class ChannelT(AsyncIterator[_EventT[_T]]):
    app: _AppT
    schema: _SchemaT
    key_type: Optional[_ModelArg]
    value_type: Optional[_ModelArg]
    loop: Optional[asyncio.AbstractEventLoop]
    maxsize: Optional[int]
    active_partitions: Optional[Set[TP]]

    @abc.abstractmethod
    def __init__(self,
                 app: _AppT,
                 *,
                 schema: _SchemaT = None,
                 key_type: _ModelArg = None,
                 value_type: _ModelArg = None,
                 is_iterator: bool = False,
                 queue: ThrowableQueue = None,
                 maxsize: int = None,
                 root: 'ChannelT' = None,
                 active_partitions: Set[TP] = None,
                 loop: asyncio.AbstractEventLoop = None) -> None:
        ...

    @abc.abstractmethod
    def clone(self, *, is_iterator: bool = None,
              **kwargs: Any) -> 'ChannelT[_T]':
        ...

    @abc.abstractmethod
    def clone_using_queue(self, queue: asyncio.Queue) -> 'ChannelT[_T]':
        ...

    @abc.abstractmethod
    def stream(self, **kwargs: Any) -> '_StreamT[_T]':
        ...

    @abc.abstractmethod
    def get_topic_name(self) -> str:
        ...

    @abc.abstractmethod
    async def send(self,
                   *,
                   key: K = None,
                   value: V = None,
                   partition: int = None,
                   timestamp: float = None,
                   headers: HeadersArg = None,
                   schema: _SchemaT = None,
                   key_serializer: CodecArg = None,
                   value_serializer: CodecArg = None,
                   callback: MessageSentCallback = None,
                   force: bool = False) -> Awaitable[RecordMetadata]:
        ...

    @abc.abstractmethod
    def send_soon(self,
                  *,
                  key: K = None,
                  value: V = None,
                  partition: int = None,
                  timestamp: float = None,
                  headers: HeadersArg = None,
                  schema: _SchemaT = None,
                  key_serializer: CodecArg = None,
                  value_serializer: CodecArg = None,
                  callback: MessageSentCallback = None,
                  force: bool = False,
                  eager_partitioning: bool = False) -> FutureMessage:
        ...

    @abc.abstractmethod
    def as_future_message(
            self,
            key: K = None,
            value: V = None,
            partition: int = None,
            timestamp: float = None,
            headers: HeadersArg = None,
            schema: _SchemaT = None,
            key_serializer: CodecArg = None,
            value_serializer: CodecArg = None,
            callback: MessageSentCallback = None,
            eager_partitioning: bool = False) -> FutureMessage:
        ...

    @abc.abstractmethod
    async def publish_message(self, fut: FutureMessage,
                              wait: bool = True) -> Awaitable[RecordMetadata]:
        ...

    @stampede
    @abc.abstractmethod
    async def maybe_declare(self) -> None:
        ...

    @abc.abstractmethod
    async def declare(self) -> None:
        ...

    @abc.abstractmethod
    def prepare_key(self,
                    key: K,
                    key_serializer: CodecArg,
                    schema: _SchemaT = None) -> Any:
        ...

    @abc.abstractmethod
    def prepare_value(self,
                      value: V,
                      value_serializer: CodecArg,
                      schema: _SchemaT = None) -> Any:
        ...

    @abc.abstractmethod
    async def decode(self, message: Message, *,
                     propagate: bool = False) -> _EventT[_T]:
        ...

    @abc.abstractmethod
    async def deliver(self, message: Message) -> None:
        ...

    @abc.abstractmethod
    async def put(self, value: _EventT[_T]) -> None:
        ...

    @abc.abstractmethod
    async def get(self, *, timeout: Seconds = None) -> _EventT[_T]:
        ...

    @abc.abstractmethod
    def empty(self) -> bool:
        ...

    @abc.abstractmethod
    async def on_key_decode_error(self, exc: Exception,
                                  message: Message) -> None:
        ...

    @abc.abstractmethod
    async def on_value_decode_error(self, exc: Exception,
                                    message: Message) -> None:
        ...

    @abc.abstractmethod
    async def on_decode_error(self, exc: Exception, message: Message) -> None:
        ...

    @abc.abstractmethod
    def on_stop_iteration(self) -> None:
        ...

    @abc.abstractmethod
    def __aiter__(self) -> 'ChannelT':
        ...

    @abc.abstractmethod
    def __anext__(self) -> Awaitable[_EventT[_T]]:
        ...

    @abc.abstractmethod
    async def throw(self, exc: BaseException) -> None:
        ...

    @abc.abstractmethod
    def _throw(self, exc: BaseException) -> None:
        ...

    @abc.abstractmethod
    def derive(self, **kwargs: Any) -> 'ChannelT':
        ...

    @property
    @abc.abstractmethod
    def subscriber_count(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def queue(self) -> ThrowableQueue:
        ...



from mode.utils.mocks import Mock
from faust.livecheck.locals import current_execution, current_execution_stack


def test_current_execution():
    m1 = Mock(name='m1')
    assert current_execution() is None
    with current_execution_stack.push(m1):
        assert current_execution() is m1
        m2 = Mock(name='m2')
        with current_execution_stack.push(m2):
            assert current_execution() is m2
            m3 = Mock(name='m3')
            with current_execution_stack.push(m3):
                assert current_execution() is m3
            assert current_execution() is m2
        assert current_execution() is m1
    assert current_execution() is None



import asyncio

import faust
import pytest
from faust import App, Channel, Record
from faust.agents.actor import Actor

from faust.agents.models import (
    ModelReqRepRequest,
    ModelReqRepResponse,
    ReqRepRequest,
    ReqRepResponse,
)
from faust.agents.replies import ReplyConsumer
from faust.events import Event
from faust.exceptions import ImproperlyConfigured
from faust.types import Message, TP
from mode import SupervisorStrategy, label
from mode.utils.aiter import aiter
from mode.utils.futures import done_future
from mode.utils.logging import CompositeLogger
from mode.utils.mocks import ANY, AsyncMock, FutureMock, Mock, call, patch
from mode.utils.trees import Node


class Word(Record):
    word: str


class test_AgentService:

    @pytest.fixture
    def agent(self, *, app):

        @app.agent()
        async def myagent(stream):
            async for value in stream:
                yield value

        return myagent

    @pytest.mark.asyncio
    @pytest.mark.parametrize('concurrency,index,expected_index', [
        (1, 3, None),
        (10, 3, 3),
    ])
    async def test_start_one(self, concurrency, index, expected_index, *,
                             agent):
        agent.concurrency = concurrency
        agent._start_task = AsyncMock(name='_start_task')

        tps = {TP('foo', 0)}
        await agent._start_one(index=index, active_partitions=tps)
        agent._start_task.assert_called_once_with(
            index=expected_index,
            active_partitions=tps,
            stream=None,
            channel=None,
            beacon=agent.beacon,
        )

    @pytest.mark.asyncio
    async def test_start_for_partitions(self, *, agent):
        agent._start_one_supervised = AsyncMock(name='_start_one')
        s = agent._start_one_supervised.return_value = Mock(name='service')
        s.maybe_start = AsyncMock(name='agent.maybe_start')
        tps = {TP('foo', 3)}
        await agent._start_for_partitions(tps)
        agent._start_one_supervised.assert_called_once_with(None, tps)

    @pytest.mark.asyncio
    async def test_on_start(self, *, agent):
        agent._new_supervisor = Mock(name='new_supervisor')
        agent._on_start_supervisor = AsyncMock(name='on_start_supervisor')
        await agent.on_start()

        agent._new_supervisor.assert_called_once_with()
        assert agent.supervisor is agent._new_supervisor()
        agent._on_start_supervisor.assert_called_once_with()

    def test_new_supervisor(self, *, agent):
        strategy = agent._get_supervisor_strategy = Mock(name='strategy')
        s = agent._new_supervisor()
        strategy.assert_called_once_with()
        strategy.return_value.assert_called_once_with(
            max_restarts=100.0,
            over=1.0,
            replacement=agent._replace_actor,
            loop=agent.loop,
            beacon=agent.beacon,
        )
        assert s is strategy()()

    def test_get_supervisor_strategy(self, *, agent):
        agent.supervisor_strategy = 100
        assert agent._get_supervisor_strategy() == 100
        agent.supervisor_strategy = None
        agent.app = Mock(name='app', autospec=App)
        assert (agent._get_supervisor_strategy() is
                agent.app.conf.agent_supervisor)

    @pytest.mark.asyncio
    async def test_on_start_supervisor(self, *, agent):
        agent.concurrency = 10
        agent._get_active_partitions = Mock(name='_get_active_partitions')
        agent._start_one = AsyncMock(name='_start_one')
        agent.supervisor = Mock(
            name='supervisor',
            autospec=SupervisorStrategy,
            start=AsyncMock(),
        )
        await agent._on_start_supervisor()

        aref = agent._start_one.coro.return_value

        agent._start_one.coro.assert_has_calls([
            call(index=i,
                 channel=aref.stream.channel if i else None,
                 active_partitions=agent._get_active_partitions())
            for i in range(10)
        ])
        agent.supervisor.add.assert_has_calls([
            call(agent._start_one.coro()) for i in range(10)
        ])
        agent.supervisor.start.assert_called_once_with()

    def test_get_active_partitions(self, *, agent):
        agent.isolated_partitions = None
        assert agent._get_active_partitions() is None
        agent.isolated_partitions = True
        assert agent._get_active_partitions() == set()
        assert agent._pending_active_partitions == set()

    @pytest.mark.asyncio
    async def test_replace_actor(self, *, agent):
        aref = Mock(name='aref', autospec=Actor)
        agent._start_one = AsyncMock(name='_start_one')
        assert (await agent._replace_actor(aref, 101) ==
                agent._start_one.coro())
        agent._start_one.assert_called_once_with(
            index=101,
            active_partitions=aref.active_partitions,
            stream=aref.stream,
            channel=aref.stream.channel,
        )

    @pytest.mark.asyncio
    async def test_on_stop(self, *, agent):
        agent._stop_supervisor = AsyncMock(name='_stop_supervisor')
        await agent.on_stop()
        agent._stop_supervisor.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_stop_supervisor(self, *, agent):
        supervisor = agent.supervisor = Mock(
            name='supervisor',
            autospec=SupervisorStrategy,
            stop=AsyncMock(),
        )
        await agent._stop_supervisor()
        assert agent.supervisor is None
        await agent._stop_supervisor()
        supervisor.stop.assert_called_once_with()

    def test_label(self, *, agent):
        assert label(agent)


class test_Agent:

    @pytest.fixture
    def agent(self, *, app):

        @app.agent()
        async def myagent(stream):
            async for value in stream:
                yield value

        return myagent

    @pytest.fixture
    def isolated_agent(self, *, app):

        @app.agent(isolated_partitions=True)
        async def isoagent(stream):
            async for value in stream:
                yield value

        return isoagent

    @pytest.fixture
    def foo_topic(self, *, app):
        return app.topic('foo')

    @pytest.fixture
    def agent2(self, *, app, foo_topic):

        @app.agent(foo_topic)
        async def other_agent(stream):
            async for value in stream:
                value

        return other_agent

    def test_init_schema_and_channel(self, *, app):
        with pytest.raises(AssertionError):
            @app.agent(app.topic('foo'), schema=faust.Schema(key_type=bytes))
            async def foo():
                ...

    def test_init_key_type_and_channel(self, *, app):
        with pytest.raises(AssertionError):
            @app.agent(app.topic('foo'), key_type=bytes)
            async def foo():
                ...

    def test_init_value_type_and_channel(self, *, app):
        with pytest.raises(AssertionError):
            @app.agent(app.topic('foo'), value_type=bytes)
            async def foo():
                ...

    def test_isolated_partitions_cannot_have_concurrency(self, *, app):
        with pytest.raises(ImproperlyConfigured):
            @app.agent(isolated_partitions=True, concurrency=100)
            async def foo():
                ...

    def test_agent_call_reuse_stream(self, *, agent, app):
        stream = app.stream('foo')
        stream.concurrency_index = 1
        stream.active_partitions = {1, 2}
        actor = agent(stream=stream, index=1, active_partitions={1, 2})
        assert actor.stream is stream

    def test_cancel(self, *, agent):
        actor1 = Mock(name='actor1')
        actor2 = Mock(name='actor2')
        agent._actors = [actor1, actor2]
        agent.cancel()
        actor1.cancel.assert_called_once_with()
        actor2.cancel.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_partitions_revoked(self, *, agent):
        revoked = {TP('foo', 0)}
        agent.on_shared_partitions_revoked = AsyncMock(name='ospr')
        await agent.on_partitions_revoked(revoked)
        agent.on_shared_partitions_revoked.assert_called_once_with(revoked)

    @pytest.mark.asyncio
    async def test_on_partitions_revoked__isolated(self, *, isolated_agent):
        revoked = {TP('foo', 0)}
        i = isolated_agent.on_isolated_partitions_revoked = AsyncMock(name='i')
        await isolated_agent.on_partitions_revoked(revoked)
        i.assert_called_once_with(revoked)

    @pytest.mark.asyncio
    async def test_on_partitions_assigned(self, *, agent):
        assigned = {TP('foo', 0)}
        agent.on_shared_partitions_assigned = AsyncMock(name='ospr')
        await agent.on_partitions_assigned(assigned)
        agent.on_shared_partitions_assigned.assert_called_once_with(assigned)

    @pytest.mark.asyncio
    async def test_on_partitions_assigned__isolated(self, *, isolated_agent):
        assigned = {TP('foo', 0)}
        i = isolated_agent.on_isolated_partitions_assigned = AsyncMock()
        await isolated_agent.on_partitions_assigned(assigned)
        i.assert_called_once_with(assigned)

    @pytest.mark.asyncio
    async def test_on_isolated_partitions_revoked(self, *, agent):
        tp = TP('foo', 0)
        aref = Mock(
            name='aref',
            autospec=Actor,
            on_isolated_partition_revoked=AsyncMock(),
        )
        agent._actor_by_partition = {tp: aref}

        await agent.on_isolated_partitions_revoked({tp})
        aref.on_isolated_partition_revoked.assert_called_once_with(tp)
        assert not agent._actor_by_partition
        await agent.on_isolated_partitions_revoked({tp})

    @pytest.mark.asyncio
    async def test_on_isolated_partitions_assigned(self, *, agent):
        agent._assign_isolated_partition = AsyncMock(name='aip')
        await agent.on_isolated_partitions_assigned({TP('foo', 0)})
        agent._assign_isolated_partition.assert_called_once_with(TP('foo', 0))

    @pytest.mark.asyncio
    async def test_assign_isolated_partition(self, *, agent):
        agent._on_first_isolated_partition_assigned = Mock(name='ofipa')
        agent._maybe_start_isolated = AsyncMock(name='maybe_start_isolated')
        agent._first_assignment_done = True

        tp = TP('foo', 606)
        await agent._assign_isolated_partition(tp)

        agent._on_first_isolated_partition_assigned.assert_not_called()
        agent._maybe_start_isolated.assert_called_once_with(tp)

        agent._first_assignment_done = False
        agent._actor_by_partition = set()
        await agent._assign_isolated_partition(tp)
        agent._on_first_isolated_partition_assigned.assert_called_once_with(tp)

    def test_on_first_isolated_partition_assigned(self, *, agent):
        aref = Mock(name='actor', autospec=Actor)
        agent._actors = [aref]
        agent._pending_active_partitions = set()
        tp = TP('foo', 303)
        agent._on_first_isolated_partition_assigned(tp)
        assert agent._actor_by_partition[tp] is aref
        assert agent._pending_active_partitions == {tp}
        agent._pending_active_partitions = None
        agent._on_first_isolated_partition_assigned(tp)

    @pytest.mark.asyncio
    async def test_maybe_start_isolated(self, *, isolated_agent):
        aref = Mock(
            name='actor',
            autospec=Actor,
            on_isolated_partition_assigned=AsyncMock(),
        )
        isolated_agent._start_isolated = AsyncMock(
            name='_start_isolated',
            return_value=aref,
        )
        tp = TP('foo', 303)
        await isolated_agent._maybe_start_isolated(tp)

        isolated_agent._start_isolated.assert_called_once_with(tp)
        assert isolated_agent._actor_by_partition[tp] is aref
        aref.on_isolated_partition_assigned.assert_called_once_with(tp)

    @pytest.mark.asyncio
    async def test_start_isolated(self, *, agent):
        agent._start_for_partitions = AsyncMock(
            name='agent._start_for_partitions',
        )
        ret = await agent._start_isolated(TP('foo', 0))
        agent._start_for_partitions.assert_called_once_with({TP('foo', 0)})
        assert ret is agent._start_for_partitions.coro()

    @pytest.mark.asyncio
    async def test_on_shared_partitions_revoked(self, *, agent):
        await agent.on_shared_partitions_revoked(set())

    @pytest.mark.asyncio
    async def test_on_shared_partitions_assigned(self, *, agent):
        await agent.on_shared_partitions_assigned(set())

    def test_info(self, *, agent):
        assert agent.info() == {
            'app': agent.app,
            'fun': agent.fun,
            'name': agent.name,
            'channel': agent.channel,
            'concurrency': agent.concurrency,
            'help': agent.help,
            'sink': agent._sinks,
            'on_error': agent._on_error,
            'supervisor_strategy': agent.supervisor_strategy,
            'isolated_partitions': agent.isolated_partitions,
        }

    def test_clone(self, *, agent):
        assert agent.clone(isolated_partitions=True).isolated_partitions

    def test_stream__active_partitions(self, *, agent):
        assert agent.stream(active_partitions={TP('foo', 0)})

    @pytest.mark.parametrize('input,expected', [
        (ReqRepRequest('value', 'reply_to', 'correlation_id'), 'value'),
        ('value', 'value'),
    ])
    def test_maybe_unwrap_reply_request(self, input, expected, *, agent):
        assert agent._maybe_unwrap_reply_request(input) == expected

    @pytest.mark.asyncio
    async def test_start_task(self, *, agent):
        agent._prepare_actor = AsyncMock(name='_prepare_actor')
        ret = await agent._start_task(index=0)
        agent._prepare_actor.assert_called_once_with(ANY, agent.beacon)
        assert ret is agent._prepare_actor.coro()

    @pytest.mark.asyncio
    async def test_prepare_actor__AsyncIterable(self, *, agent):
        aref = agent(index=0, active_partitions=None)
        with patch('asyncio.Task') as Task:
            agent._slurp = Mock(name='_slurp')
            agent._execute_actor = Mock(name='_execute_actor')
            beacon = Mock(name='beacon', autospec=Node)
            ret = await agent._prepare_actor(aref, beacon)
            agent._slurp.assert_called()
            coro = agent._slurp()
            agent._execute_actor.assert_called_once_with(coro, aref)
            Task.assert_called_once_with(
                agent._execute_actor(), loop=agent.loop)
            task = Task()
            assert task._beacon is beacon
            assert aref.actor_task is task
            assert aref in agent._actors
            assert ret is aref

    @pytest.mark.asyncio
    async def test_prepare_actor__Awaitable(self, *, agent2):
        aref = agent2(index=0, active_partitions=None)
        asyncio.ensure_future(aref.it).cancel()  # silence warning
        return
        with patch('asyncio.Task') as Task:
            agent2._execute_actor = Mock(name='_execute_actor')
            beacon = Mock(name='beacon', autospec=Node)
            ret = await agent2._prepare_actor(aref, beacon)
            coro = aref
            agent2._execute_actor.assert_called_once_with(coro, aref)
            Task.assert_called_once_with(
                agent2._execute_actor(), loop=agent2.loop)
            task = Task()
            assert task._beacon is beacon
            assert aref.actor_task is task
            assert aref in agent2._actors
            assert ret is aref

    @pytest.mark.asyncio
    async def test_prepare_actor__Awaitable_cannot_have_sinks(self, *, agent2):
        aref = agent2(index=0, active_partitions=None)
        asyncio.ensure_future(aref.it).cancel()  # silence warning
        agent2._sinks = [agent2]
        with pytest.raises(ImproperlyConfigured):
            await agent2._prepare_actor(
                aref,
                Mock(name='beacon', autospec=Node),
            )

    @pytest.mark.asyncio
    async def test_execute_actor(self, *, agent):
        coro = done_future()
        await agent._execute_actor(coro, Mock(name='aref', autospec=Actor))

    @pytest.mark.asyncio
    async def test_execute_actor__cancelled_stopped(self, *, agent):
        coro = FutureMock()
        coro.side_effect = asyncio.CancelledError()
        await agent.stop()
        with pytest.raises(asyncio.CancelledError):
            await agent._execute_actor(coro, Mock(name='aref', autospec=Actor))
        coro.assert_awaited()

    @pytest.mark.asyncio
    async def test_execute_actor__cancelled_running(self, *, agent):
        coro = FutureMock()
        coro.side_effect = asyncio.CancelledError()
        await agent._execute_actor(coro, Mock(name='aref', autospec=Actor))
        coro.assert_awaited()

    @pytest.mark.asyncio
    async def test_execute_actor__raising(self, *, agent):
        agent._on_error = AsyncMock(name='on_error')
        agent.log = Mock(name='log', autospec=CompositeLogger)
        aref = Mock(
            name='aref',
            autospec=Actor,
            crash=AsyncMock(),
        )
        agent.supervisor = Mock(name='supervisor')
        coro = FutureMock()
        exc = coro.side_effect = KeyError('bar')
        await agent._execute_actor(coro, aref)
        coro.assert_awaited()

        aref.crash.assert_called_once_with(exc)
        agent.supervisor.wakeup.assert_called_once_with()
        agent._on_error.assert_called_once_with(agent, exc)

        agent._on_error = None
        await agent._execute_actor(coro, aref)

    @pytest.mark.asyncio
    async def test_slurp(self, *, agent, app):
        aref = agent(index=None, active_partitions=None)
        stream = aref.stream.get_active_stream()
        agent._delegate_to_sinks = AsyncMock(name='_delegate_to_sinks')
        agent._reply = AsyncMock(name='_reply')

        def on_delegate(value):
            raise StopAsyncIteration()

        word = Word('word')
        word_req = ReqRepRequest(word, 'reply_to', 'correlation_id')
        message1 = Mock(name='message1', autospec=Message)
        message2 = Mock(name='message2', autospec=Message)
        event1 = Event(app, None, word_req, {}, message1)
        event2 = Event(app, 'key', 'bar', {}, message2)
        values = [
            (event1, word),
            (event2, 'bar'),
        ]

        class AIT:

            async def __aiter__(self):
                for event, value in values:
                    stream.current_event = event
                    yield value
        it = aiter(AIT())
        await agent._slurp(aref, it)

        agent._reply.assert_called_once_with(
            None, word, word_req.reply_to, word_req.correlation_id)
        agent._delegate_to_sinks.coro.assert_has_calls([
            call(word),
            call('bar'),
        ])

    @pytest.mark.asyncio
    async def test_slurp__headers(self, *, agent, app):
        agent.use_reply_headers = True
        aref = agent(index=None, active_partitions=None)
        stream = aref.stream.get_active_stream()
        agent._delegate_to_sinks = AsyncMock(name='_delegate_to_sinks')
        agent._reply = AsyncMock(name='_reply')

        def on_delegate(value):
            raise StopAsyncIteration()

        word = Word('word')
        message1 = Mock(name='message1', autospec=Message)
        headers1 = message1.headers = {
            'Faust-Ag-ReplyTo': 'reply_to',
            'Faust-Ag-CorrelationId': 'correlation_id',
        }
        message2 = Mock(name='message2', autospec=Message)
        headers2 = message2.headers = {}
        event1 = Event(app, None, word, headers1, message1)
        event2 = Event(app, 'key', 'bar', headers2, message2)
        values = [
            (event1, word, True),
            (event1, word, False),
            (event2, 'bar', True),
        ]

        class AIT:

            async def __aiter__(self):
                for event, value, set_cur_event in values:
                    if set_cur_event:
                        stream.current_event = event
                    else:
                        stream.current_event = None
                    yield value
        it = aiter(AIT())
        await agent._slurp(aref, it)

        agent._reply.assert_called_once_with(
            None, word, 'reply_to', 'correlation_id')
        agent._delegate_to_sinks.coro.assert_has_calls([
            call(word),
            call('bar'),
        ])

    @pytest.mark.asyncio
    async def test_delegate_to_sinks(self, *, agent, agent2, foo_topic):
        agent2.send = AsyncMock(name='agent2.send')
        foo_topic.send = AsyncMock(name='foo_topic.send')
        sink_callback = Mock(name='sink_callback')
        sink_callback2_mock = Mock(name='sink_callback2_mock')

        async def sink_callback2(value):
            return sink_callback2_mock(value)

        agent._sinks = [
            agent2,
            foo_topic,
            sink_callback,
            sink_callback2,
        ]

        value = Mock(name='value')
        await agent._delegate_to_sinks(value)

        agent2.send.assert_called_once_with(value=value)
        foo_topic.send.assert_called_once_with(value=value)
        sink_callback.assert_called_once_with(value)
        sink_callback2_mock.assert_called_once_with(value)

    @pytest.mark.asyncio
    async def test_reply(self, *, agent):
        agent.app = Mock(
            name='app',
            autospec=App,
            send=AsyncMock(),
        )
        req = ReqRepRequest('value', 'reply_to', 'correlation_id')
        await agent._reply('key', 'reply', req.reply_to, req.correlation_id)
        agent.app.send.assert_called_once_with(
            req.reply_to,
            key=None,
            value=ReqRepResponse(
                key='key',
                value='reply',
                correlation_id=req.correlation_id,
            ),
        )

    @pytest.mark.asyncio
    async def test_cast(self, *, agent):
        agent.send = AsyncMock(name='send')
        await agent.cast('value', key='key', partition=303)
        agent.send.assert_called_once_with(
            key='key',
            value='value',
            partition=303,
            headers=None,
            timestamp=None,
        )

    @pytest.mark.asyncio
    async def test_ask(self, *, agent):
        agent.app = Mock(
            name='app',
            autospec=App,
            maybe_start_client=AsyncMock(),
            _reply_consumer=Mock(
                autospec=ReplyConsumer,
                add=AsyncMock(),
            ),
        )
        pp = done_future()
        agent.ask_nowait = Mock(name='ask_nowait')
        agent.ask_nowait.return_value = done_future(pp)
        pp.correlation_id = 'foo'

        await agent.ask(
            value='val',
            key='key',
            partition=303,
            correlation_id='correlation_id',
            headers={'k1': 'v1'},
        )
        agent.ask_nowait.assert_called_once_with(
            'val',
            key='key',
            partition=303,
            reply_to=agent.app.conf.reply_to,
            correlation_id='correlation_id',
            force=True,
            timestamp=None,
            headers={'k1': 'v1'},
        )
        agent.app._reply_consumer.add.assert_called_once_with(
            pp.correlation_id, pp)

    @pytest.mark.asyncio
    async def test_ask_nowait(self, *, agent):
        agent._create_req = Mock(name='_create_req')
        agent._create_req.return_value = ['V', None]
        agent.channel.send = AsyncMock(name='channel.send')
        res = await agent.ask_nowait(
            value='value',
            key='key',
            partition=303,
            timestamp=None,
            headers=None,
            reply_to='reply_to',
            correlation_id='correlation_id',
            force=True,
        )

        agent._create_req.assert_called_once_with(
            'key', 'value', 'reply_to', 'correlation_id', None)
        agent.channel.send.assert_called_once_with(
            key='key',
            value=agent._create_req()[0],
            partition=303,
            timestamp=None,
            force=True,
            headers=None,
        )

        assert res.reply_to
        assert res.correlation_id

    @pytest.mark.asyncio
    async def test_ask_nowait__missing_reply_to(self, *, agent):
        with pytest.raises(TypeError):
            await agent.ask_nowait(
                value='value',
                key='key',
                partition=3034,
                reply_to=None,
            )

    def test_create_req(self, *, agent):
        agent.use_reply_headers = False
        agent._get_strtopic = Mock(name='_get_strtopic')
        with patch('faust.agents.agent.uuid4') as uuid4:
            uuid4.return_value = 'vvv'
            reqrep = agent._create_req(
                key=b'key', value=b'value',
                reply_to='reply_to',
                headers={'k': 'v'})[0]

            agent._get_strtopic.assert_called_once_with('reply_to')

            assert reqrep.value == b'value'
            assert reqrep.reply_to == agent._get_strtopic()
            assert reqrep.correlation_id == 'vvv'

    def test_create_req__use_reply_headers(self, *, agent):
        agent.use_reply_headers = True
        agent._get_strtopic = Mock(name='_get_strtopic')
        with patch('faust.agents.agent.uuid4') as uuid4:
            uuid4.return_value = 'vvv'
            value, h = agent._create_req(
                key=b'key', value=b'value',
                reply_to='reply_to',
                headers={'k': 'v'})

            agent._get_strtopic.assert_called_once_with('reply_to')

            assert value == b'value'
            assert h['Faust-Ag-ReplyTo'] == agent._get_strtopic()
            assert h['Faust-Ag-CorrelationId'] == 'vvv'.encode()

    def test_create_req__model(self, *, agent):
        agent.use_reply_headers = False
        agent._get_strtopic = Mock(name='_get_strtopic')
        with patch('faust.agents.agent.uuid4') as uuid4:
            uuid4.return_value = 'vvv'
            value = Word('foo')
            reqrep = agent._create_req(
                key=b'key',
                value=value,
                reply_to='reply_to',
                headers={'h1': 'h2'})[0]
            assert isinstance(reqrep, ReqRepRequest)

            agent._get_strtopic.assert_called_once_with('reply_to')
            assert isinstance(reqrep, ModelReqRepRequest)

            assert reqrep.value is value
            assert reqrep.reply_to == agent._get_strtopic()
            assert reqrep.correlation_id == 'vvv'

    def test_create_req__requires_reply_to(self, *, agent):
        with pytest.raises(TypeError):
            agent._create_req(
                key=b'key', value=b'value', reply_to=None,
            )

    @pytest.mark.parametrize('value,expected_class', [
        (b'value', ReqRepResponse),
        (Word('foo'), ModelReqRepResponse),
    ])
    def test_response_class(self, value, expected_class, *, agent):
        assert agent._response_class(value) is expected_class

    @pytest.mark.asyncio
    async def test_send(self, *, agent):
        agent.channel = Mock(
            name='channel',
            autospec=Channel,
            send=AsyncMock(),
        )
        agent._create_req = Mock(name='_create_req')
        agent._create_req.return_value = ('V', {'k2': 'v2'})
        callback = Mock(name='callback')

        ret = await agent.send(
            key=b'key',
            value=b'value',
            partition=303,
            timestamp=None,
            headers={'k': 'v'},
            key_serializer='raw',
            value_serializer='raw',
            callback=callback,
            reply_to='reply_to',
            correlation_id='correlation_id',
            force=True,
        )

        agent._create_req.assert_called_once_with(
            b'key', b'value', 'reply_to', 'correlation_id', {'k': 'v'},
        )

        agent.channel.send.assert_called_once_with(
            key=b'key',
            value=agent._create_req()[0],
            partition=303,
            timestamp=None,
            headers={'k2': 'v2'},
            key_serializer='raw',
            value_serializer='raw',
            force=True,
        )

        assert ret is agent.channel.send.coro()

    @pytest.mark.asyncio
    async def test_send__without_reply_to(self, *, agent):
        agent.channel = Mock(
            name='channel',
            autospec=Channel,
            send=AsyncMock(),
        )
        agent._create_req = Mock(name='_create_req')
        callback = Mock(name='callback')

        ret = await agent.send(
            key=b'key',
            value=b'value',
            partition=303,
            timestamp=None,
            headers={'k': 'v'},
            key_serializer='raw',
            value_serializer='raw',
            callback=callback,
            reply_to=None,
            correlation_id='correlation_id',
            force=True,
        )

        agent._create_req.assert_not_called()

        agent.channel.send.assert_called_once_with(
            key=b'key',
            value=b'value',
            partition=303,
            timestamp=None,
            headers={'k': 'v'},
            key_serializer='raw',
            value_serializer='raw',
            force=True,
        )

        assert ret is agent.channel.send.coro()

    def test_get_strtopic__agent(self, *, agent, agent2):
        assert agent._get_strtopic(agent2) == agent2.channel.get_topic_name()

    def test_get_strtopic__topic(self, *, agent, foo_topic):
        assert agent._get_strtopic(foo_topic) == foo_topic.get_topic_name()

    def test_get_strtopic__str(self, *, agent):
        assert agent._get_strtopic('bar') == 'bar'

    def test_get_strtopic__channel_raises(self, *, agent, app):
        with pytest.raises(ValueError):
            agent._get_strtopic(app.channel())

    def test_get_topic_names(self, *, agent, app):
        agent.channel = app.topic('foo')
        assert agent.get_topic_names() == ('foo',)

    def test_get_topic_names__channel(self, *, agent, app):
        agent.channel = app.channel()
        assert agent.get_topic_names() == []

    def test_repr(self, *, agent):
        assert repr(agent)

    def test_channel(self, *, agent):
        agent._prepare_channel = Mock(name='_prepare_channel')
        agent._channel = None
        channel = agent.channel
        agent._prepare_channel.assert_called_once_with(
            agent._channel_arg,
            schema=agent._schema,
            key_type=agent._key_type,
            value_type=agent._value_type,
            **agent._channel_kwargs)
        assert channel is agent._prepare_channel.return_value
        assert agent._channel is channel

    def test_prepare_channel__not_channel(self, *, agent):
        with pytest.raises(TypeError):
            agent._prepare_channel(object())

    def test_add_sink(self, *, agent, agent2):
        agent.add_sink(agent2)
        assert agent2 in agent._sinks
        agent.add_sink(agent2)

    def test_channel_iterator(self, *, agent):
        agent.channel = Mock(name='channel', autospec=Channel)
        agent._channel_iterator = None
        it = agent.channel_iterator

        agent.channel.clone.assert_called_once_with(is_iterator=False)
        assert it is agent.channel.clone()
        agent.channel_iterator = [42]
        assert agent.channel_iterator == [42]

    def test_label(self, *, agent):
        assert label(agent)

    async def test_context_calls_sink(self, *, agent):
        class SinkCalledException(Exception):
            pass

        def dummy_sink(_):
            raise SinkCalledException()

        agent.add_sink(dummy_sink)
        async with agent.test_context() as agent_mock:
            with pytest.raises(SinkCalledException):
                await agent_mock.put('hello')



# cython: language_level=3
from asyncio import ALL_COMPLETED, wait
from faust.exceptions import KeyDecodeError, ValueDecodeError


cdef class ConductorHandler:

    cdef public:
        object conductor
        object tp
        object channels
        object app
        object topic
        object partition
        object on_topic_buffer_full
        object acquire_flow_control
        object consumer
        object wait_until_producer_ebb

    def __init__(self, object conductor, object tp, object channels):
        self.conductor = conductor
        self.tp = tp
        self.channels = channels

        self.app = self.conductor.app
        self.topic, self.partition = self.tp
        self.on_topic_buffer_full = self.app.sensors.on_topic_buffer_full
        self.acquire_flow_control = self.app.flow_control.acquire
        self.wait_until_producer_ebb = self.app.producer.buffer.wait_until_ebb
        self.consumer = self.app.consumer

    async def __call__(self, object message):
        cdef:
            Py_ssize_t channels_n
            object channels
            object event
            object keyid
            object dest_event
        await self.acquire_flow_control()
        await self.wait_until_producer_ebb()
        channels = self.channels
        channels_n = len(channels)
        if channels_n:
            message.refcount += channels_n  # message.incref(n)
            event = None
            event_keyid = None

            delivered = set()
            full = []
            try:
                for chan in channels:
                    event, event_keyid = self._decode(event, chan, event_keyid)
                    if event is None:
                        event = await chan.decode(message, propagate=True)
                    if not self._put(event, chan, full):
                        continue
                    delivered.add(chan)
                if full:
                    for event, chan in full:
                        self.on_topic_buffer_full(chan)
                        await chan.put(event)
                        delivered.add(chan)
            except KeyDecodeError as exc:
                remaining = channels - delivered
                message.ack(self.consumer, n=len(remaining))
                for channel in remaining:
                    await channel.on_key_decode_error(exc, message)
                    delivered.add(channel)
            except ValueDecodeError as exc:
                remaining = channels - delivered
                message.ack(self.consumer, n=len(remaining))
                for channel in remaining:
                    await channel.on_value_decode_error(exc, message)
                    delivered.add(channel)

    cdef object _decode(self, object event, object channel, object event_keyid):
        keyid = channel.key_type, channel.value_type
        if event_keyid is None or event is None:
            return None, event_keyid
        if keyid == event_keyid:
            return event, keyid

    cdef bint _put(self,
                   object event,
                   object channel,
                   object full):
        cdef:
            object queue
        queue = channel.queue
        if queue.full():
            full.append((event, channel))
            return False
        else:
            queue.put_nowait(event)
            return True



from mode import Service, label
from mode.utils.mocks import AsyncMock, Mock, call
import pytest


class OtherService(Service):
    ...


class test_AppService:

    def test_on_init_dependencies(self, *, app):
        app.boot_strategy = Mock(name='boot_strategy')
        app.client_only = True
        assert app.on_init_dependencies() == app.boot_strategy.client_only()

        app.client_only = False
        assert app.on_init_dependencies() == app.boot_strategy.server()

        app.producer_only = True
        assert app.on_init_dependencies() == app.boot_strategy.producer_only()

    def test_components_client(self, *, app):
        assert list(app.boot_strategy.client_only()) == [
            app.producer,
            app.consumer,
            app._reply_consumer,
            app.topics,
            app._fetcher,
        ]

    def test_components_producer_only(self, *, app):
        assert list(app.boot_strategy.producer_only()) == [
            app.cache,
            app.web,
            app.producer,
        ]

    def test_components_server(self, *, app):
        components = list(app.boot_strategy.server())
        expected_components = list(app.sensors)
        expected_components.extend([
            app.producer,
            app.cache,
            app.web,
            app.consumer,
            app._leader_assignor,
            app._reply_consumer,
        ])
        expected_components.extend(list(app.agents.values()))
        expected_components.extend([
            app.agents,
            app.topics,
            app.tables,
        ])
        assert components == expected_components

    @pytest.mark.asyncio
    async def test_on_first_start(self, *, app):
        app.agents = Mock(name='agents')
        app._create_directories = Mock(name='app._create_directories')
        await app.on_first_start()

        app._create_directories.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_start(self, *, app):
        app.finalize = Mock(name='app.finalize')
        await app.on_start()

        app.finalize.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_started(self, *, app):
        app._wait_for_table_recovery_completed = AsyncMock(return_value=True)
        app.on_started_init_extra_tasks = AsyncMock(name='osiet')
        app.on_started_init_extra_services = AsyncMock(name='osies')
        app.on_startup_finished = None
        app._wait_for_table_recovery_completed.coro.return_value = True
        await app.on_started()

        app._wait_for_table_recovery_completed.coro.return_value = False
        await app.on_started()

        app.on_started_init_extra_tasks.assert_called_once_with()
        app.on_started_init_extra_services.assert_called_once_with()

        app.on_startup_finished = AsyncMock(name='on_startup_finished')
        await app.on_started()

        app.on_startup_finished.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_wait_for_table_recovery_completed(self, *, app):
        app.tables = Mock()
        app.tables.recovery.started = True
        app.tables.wait_until_recovery_completed = AsyncMock()
        await app._wait_for_table_recovery_completed()
        app.tables.wait_until_recovery_completed.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_started_init_extra_tasks(self, *, app):
        app.add_future = Mock(name='add_future')

        t1_mock = Mock(name='t1_mock')
        t2_mock = Mock(name='t2_mock')

        def t1():
            return t1_mock()

        def t2():
            return t2_mock(self)

        app._app_tasks = [t1, t2]
        await app.on_started_init_extra_tasks()

        app.add_future.assert_has_calls([
            call(t1()),
            call(t2()),
        ])

        t1_mock.assert_called_with()
        t2_mock.assert_called_with(self)

    @pytest.mark.asyncio
    async def test_on_started_init_extra_services(self, *, app):
        app.add_runtime_dependency = AsyncMock(name='add_runtime_dependency')
        service1 = Mock(name='service1', autospec=Service)
        app._extra_services = [service1]
        app._extra_service_instances = None
        await app.on_started_init_extra_services()

        app.add_runtime_dependency.assert_called_once_with(service1)
        assert app._extra_service_instances == [service1]
        await app.on_started_init_extra_services()  # noop

    def test_prepare_subservice(self, *, app):
        service = OtherService()
        assert app._prepare_subservice(service) is service

    def test_prepare_subservice__class(self, *, app):
        service = app._prepare_subservice(OtherService)
        assert service.loop is app.loop
        assert service.beacon.parent is app.beacon

    def test_label(self, *, app):
        assert label(app)



#!/usr/bin/env python
import random
import faust


app = faust.App(
    'leader-example',
    broker='kafka://localhost:9092',
    value_serializer='raw',
)


@app.timer(2.0, on_leader=True)
async def publish_greetings():
    print('PUBLISHING ON LEADER!')
    await say.send(value=str(random.random()))


@app.agent()
async def say(greetings):
    async for greeting in greetings:
        print(greeting)



#!/usr/bin/env python
import faust
TOPIC = 'test'

# host.docker.internal is how a docker container connects to the local machine.
# Don't use in production, this only works with Docker for Mac in development
app = faust.App('scores', broker='kafka://host.docker.internal:9092')
table = app.Table('totals', default=int)


class Score(faust.Record, serializer='json'):
    index: int
    value: int


test_topic = app.topic(TOPIC, value_type=Score)


def get_score_key(score):
    return f'partition: {score.index % 2}'


@app.agent(test_topic)
async def print_totals(stream):
    async for score in stream.group_by(get_score_key, name='index_partition'):
        ind = f'partition: {score.index % 2}'
        table['totals'] += 1
        table[ind] += 1
        print(f'Total: {table["totals"]}, Partition {ind}: {table[ind]}')


if __name__ == '__main__':
    app.main()



import abc
import typing
from typing import Any, List, MutableMapping, Set

from mode import ServiceT
from yarl import URL

from .tuples import TP

if typing.TYPE_CHECKING:
    from .app import AppT as _AppT
else:
    class _AppT: ...      # noqa

__all__ = [
    'TopicToPartitionMap',
    'HostToPartitionMap',
    'PartitionAssignorT',
    'LeaderAssignorT',
]

TopicToPartitionMap = MutableMapping[str, List[int]]
HostToPartitionMap = MutableMapping[str, TopicToPartitionMap]


class PartitionAssignorT(abc.ABC):

    replicas: int
    app: _AppT

    @abc.abstractmethod
    def __init__(self, app: _AppT, replicas: int = 0) -> None:
        ...

    @abc.abstractmethod
    def group_for_topic(self, topic: str) -> int:
        ...

    @abc.abstractmethod
    def assigned_standbys(self) -> Set[TP]:
        ...

    @abc.abstractmethod
    def assigned_actives(self) -> Set[TP]:
        ...

    @abc.abstractmethod
    def is_active(self, tp: TP) -> bool:
        ...

    @abc.abstractmethod
    def is_standby(self, tp: TP) -> bool:
        ...

    @abc.abstractmethod
    def key_store(self, topic: str, key: bytes) -> URL:
        ...

    @abc.abstractmethod
    def table_metadata(self, topic: str) -> HostToPartitionMap:
        ...

    @abc.abstractmethod
    def tables_metadata(self) -> HostToPartitionMap:
        ...


class LeaderAssignorT(ServiceT):

    app: _AppT

    @abc.abstractmethod
    def __init__(self, app: _AppT, **kwargs: Any) -> None:
        ...

    @abc.abstractmethod
    def is_leader(self) -> bool:
        ...



# cython: language_level=3
from datetime import timedelta
from libc.math cimport floor
from faust.types import WindowT


cdef double want_seconds(object o, object default) except *:
    if o is None and default is not None:
        return default
    if isinstance(o, timedelta):
        return o.total_seconds()
    return o


cdef class SlidingWindow:
    cdef public:
        double before
        double after
        double expires

    def __init__(self, object before, object after, object expires):
        self.before = want_seconds(before, None)
        self.after = want_seconds(after, None)
        self.expires = want_seconds(expires, 0.0)

    cpdef object ranges(self, double timestamp):
        return [
            (timestamp - self.before, timestamp + self.after),
        ]

    def stale(self, double timestamp, double latest_timestamp):
        if not self.expires:
            return False
        when_stale = latest_timestamp - self.expires
        return timestamp <= when_stale

    def current(self, double timestamp):
        """Get the latest window range for a given timestamp."""
        return timestamp - self.before, timestamp + self.after

    def delta(self, double timestamp, object d):
        return self.current(timestamp - want_seconds(d, 0.0))

    def earliest(self, float timestamp):
        return self.current(timestamp)



cdef class HoppingWindow:

    cdef public:
        double size
        double step
        double expires

    def __init__(self, object size, object step,
                 object expires = 0.0):
        self.size = want_seconds(size, None)
        self.step = want_seconds(step, None)
        self.expires = want_seconds(expires, 0.0)

    cpdef object ranges(self, double timestamp):
        start = self._start_initial_range(timestamp)
        r = []
        for start in range(int(start), int(timestamp) + 1, int(self.step)):
            end = start + self.size - 0.1
            r.append((start, end))
        return r

    cdef double _start_initial_range(self, double timestamp):
        cdef:
            int rem
            double step
            double closest_step
        step = self.step
        rem = <int>floor(timestamp / step)
        closest_step = <double>rem * step
        return closest_step - self.size + step

    def stale(self, double timestamp, double latest_timestamp):
        cdef:
            double ts
        if not self.expires:
            return False
        ts = latest_timestamp - self.expires
        initial_start = self._start_initial_range(ts)
        when_stale = self._current_start(ts, initial_start)
        return timestamp <= when_stale

    cpdef object current(self, double timestamp):
        initial_start = self._start_initial_range(timestamp)
        start = self._current_start(timestamp, initial_start)
        end = start + self.size - 0.1
        return (start, end)

    cdef double _current_start(self, double timestamp, double start):
        cdef:
            double m
        step = self.step
        m = floor((timestamp - start) / step)
        return start + (step * m)

    cpdef delta(self, double timestamp, object d):
        return self.current(timestamp - want_seconds(d, None))

    cpdef object earliest(self, double timestamp):
        start = self._start_initial_range(timestamp)
        end = start + self.size - 0.1
        return (start, end)



"""Route messages to Faust nodes by partitioning."""
from typing import Tuple

from yarl import URL

from faust.exceptions import SameNode
from faust.types.app import AppT
from faust.types.assignor import PartitionAssignorT
from faust.types.core import K
from faust.types.router import HostToPartitionMap, RouterT
from faust.types.tables import CollectionT
from faust.types.web import Request, Response, Web
from faust.web.exceptions import ServiceUnavailable


class Router(RouterT):
    """Router for ``app.router``."""

    _assignor: PartitionAssignorT

    def __init__(self, app: AppT) -> None:
        self.app = app
        self._assignor = self.app.assignor

    def key_store(self, table_name: str, key: K) -> URL:
        """Return the URL of web server that hosts key in table."""
        table = self._get_table(table_name)
        topic = self._get_table_topic(table)
        k = self._get_serialized_key(table, key)
        return self._assignor.key_store(topic, k)

    def table_metadata(self, table_name: str) -> HostToPartitionMap:
        """Return metadata stored for table in the partition assignor."""
        table = self._get_table(table_name)
        topic = self._get_table_topic(table)
        return self._assignor.table_metadata(topic)

    def tables_metadata(self) -> HostToPartitionMap:
        """Return metadata stored for all tables in the partition assignor."""
        return self._assignor.tables_metadata()

    @classmethod
    def _get_table_topic(cls, table: CollectionT) -> str:
        return table.changelog_topic.get_topic_name()

    @classmethod
    def _get_serialized_key(cls, table: CollectionT, key: K) -> bytes:
        return table.changelog_topic.prepare_key(key, None)[0]

    def _get_table(self, name: str) -> CollectionT:
        return self.app.tables[name]

    async def route_req(self, table_name: str, key: K, web: Web,
                        request: Request) -> Response:
        """Route request to worker having key in table.

        Arguments:
            table_name: Name of the table.
            key: The key that we want.
            web: The currently sued web driver,
            request: The web request currently being served.
        """
        app = self.app
        try:
            dest_url: URL = app.router.key_store(table_name, key)
        except KeyError:
            raise ServiceUnavailable()
        dest_ident = (host, port) = self._urlident(dest_url)
        if dest_ident == self._urlident(app.conf.canonical_url):
            raise SameNode()
        routed_url = request.url.with_host(host).with_port(int(port))
        async with app.http_client.get(routed_url) as response:
            return web.text(
                await response.text(), content_type=response.content_type)

    def _urlident(self, url: URL) -> Tuple[str, int]:
        return (
            (url.host if url.scheme else url.path) or '',
            int(url.port or 80),
        )



import os
import faust
import pytest
from faust.transport.producer import Producer
from faust.utils.tracing import set_current_span
from mode.utils.mocks import AsyncMock, Mock


@pytest.fixture()
def app(event_loop, request):
    settings = request.node.get_closest_marker('conf')
    kwargs = settings.kwargs or {} if settings else {}

    os.environ.pop('F_DATADIR', None)
    os.environ.pop('FAUST_DATADIR', None)
    os.environ.pop('F_WORKDIR', None)
    os.environ.pop('FAUST_WORKDIR', None)
    instance = faust.App('testid', **kwargs)
    instance.producer = Mock(
        name='producer',
        autospec=Producer,
        maybe_start=AsyncMock(),
        start=AsyncMock(),
        send=AsyncMock(),
        send_and_wait=AsyncMock(),
    )
    instance.finalize()
    set_current_span(None)
    return instance


@pytest.fixture()
def web(app):
    return app.web


def pytest_configure(config):
    config.addinivalue_line(
        'markers', 'conf: Faust app configuration marker',
    )



import pytest


@pytest.fixture()
def table_foo(app):
    return app.Table('foo-table', help='First table.')


@pytest.fixture()
def table_bar(app):
    return app.Table('bar-table', help='Second table.')


@pytest.fixture()
def tables(table_foo, table_bar):
    return [table_foo, table_bar]


async def test_list_tables(web_client, tables):
    async with await web_client as client:
        resp = await client.get('/table/')
        assert resp.status == 200
        payload = await resp.json()
        tables = {t['name']: t for t in payload}
        assert 'foo-table' in tables
        assert tables['foo-table']['help'] == 'First table.'
        assert 'bar-table' in tables
        assert tables['bar-table']['help'] == 'Second table.'


async def test_table_detail(web_client, tables):
    async with await web_client as client:
        resp = await client.get('/table/foo-table/')
        assert resp.status == 200
        payload = await resp.json()
        assert payload['name'] == 'foo-table'
        assert payload['help'] == 'First table.'


async def test_table_detail__missing_table(web_client, tables):
    async with await web_client as client:
        resp = await client.get('/table/XUZZY-table/')
        assert resp.status == 404
        payload = await resp.json()
        assert payload == {
            'error': 'unknown table',
            'name': 'XUZZY-table',
        }


async def test_table_key(web_client, tables, table_foo, router_same):
    async with await web_client as client:
        table_foo.data.data['KEY'] = '303'
        resp = await client.get('/table/foo-table/KEY/')
        assert resp.status == 200
        payload = await resp.json()
        assert payload == '303'


async def test_table_key__missing_key(web_client,
                                      tables,
                                      table_foo,
                                      router_same):
    async with await web_client as client:
        resp = await client.get('/table/foo-table/MISSINGKEY/')
        assert resp.status == 404
        payload = await resp.json()
        assert payload == {
            'error': 'key not found',
            'table': 'foo-table',
            'key': 'MISSINGKEY',
        }



import asyncio
from collections import Counter
from itertools import cycle
from time import time
from typing import Any, List, Optional, Set
from faust.types import AppT, CodecT, EventT, Message as MessageT, StreamT, TP
from faust.types.core import HeadersArg
from mode import Service
from mode.utils.aiter import aenumerate

CURRENT_OFFSETS = Counter()


class AgentCase(Service):
    name: str

    wait_for_shutdown = True

    @classmethod
    async def run_test(cls, app: AppT, *,
                       num_messages: int = 100,
                       **kwargs: Any) -> 'AgentCase':
        case = cls(app, num_messages=num_messages, **kwargs)
        try:
            async with case:
                await case.wait(case.finished)
        except Exception as exc:
            print(f'AgentConcurrencyCase raised: {exc!r}')
            raise
        else:
            assert not case._crashed.is_set()
        assert case.processed_total == num_messages
        return case

    def __init__(self, app: AppT, *,
                 partitions: List[int] = None,
                 concurrency: int = 10,
                 value_serializer: CodecT = 'raw',
                 num_messages: int = 100,
                 isolated_partitions: bool = False,
                 name: str = None,
                 **kwargs: Any):
        self.app = app
        if name is not None:
            self.name = name
        if partitions is None:
            partitions = [0]
        self.partitions = sorted(partitions)
        self.concurrency = concurrency
        self.value_serializer = value_serializer
        self.num_messages = num_messages
        assert self.num_messages > 1
        self.isolated_partitions = isolated_partitions

        self.topic_name = self.name
        self.tps = [TP(self.topic_name, p) for p in self.partitions]
        self.next_tp = cycle(self.tps)
        self.expected_tp = cycle(self.tps)
        self.seen_offsets = set()
        self.processed_total = 0

        super().__init__(**kwargs)

        self.agent_started = asyncio.Event(loop=self.loop)
        self.agent_started_processing = asyncio.Event(loop=self.loop)
        self.agent_stopped_processing = asyncio.Event(loop=self.loop)
        self.finished = asyncio.Event(loop=self.loop)

    async def on_start(self) -> None:
        app = self.app
        topic = app.topic(self.topic_name,
                          value_serializer=self.value_serializer)
        create_agent = app.agent(
            topic,
            concurrency=self.concurrency,
            isolated_partitions=self.isolated_partitions,
        )
        self.agent = create_agent(self.process)
        assert self.agent

    @Service.task
    async def _wait_for_success(self) -> None:
        await self.wait(self.finished)
        try:
            await self.assert_success()
        finally:
            self.set_shutdown()

    async def assert_success(self) -> None:
        assert self.processed_total == self.num_messages

    async def on_agent_event(
            self, stream: StreamT, event: EventT) -> None:
        ...

    async def process(self, stream: StreamT[bytes]) -> None:
        self.agent_started.set()
        try:
            async for i, event in aenumerate(stream.events()):
                self.agent_started_processing.set()
                self.processed_total += 1

                await self.on_agent_event(stream, event)

                key = event.message.tp, event.message.offset
                if key in self.seen_offsets:
                    print(f'EVENT PROCESSED TWICE: {key}')
                    await self.crash(
                        Exception(f'Event processed twice: {key}'))
                self.seen_offsets.add(key)
                assert key in self.seen_offsets
                if self.processed_total >= self.num_messages:
                    self.agent_stopped_processing.set()
                await self.sleep(0)
        except asyncio.CancelledError:
            if self.processed_total < self.num_messages:
                print('I WAS CANCELLED?!?!?')
            raise
        except Exception as exc:
            print(f'AGENT RAISED ERROR: {exc!r}')
            await self.crash(exc)
            raise

    @Service.task
    async def _send(self) -> None:
        app = self.app

        async with app.agents:
            app.flow_control.resume()
            # Wait for agent to enter execution.
            await self.wait(self.agent_started, timeout=1.0)
            # Wait for Stream.__aiter__ to start up.
            await self.sleep(0.5)

            # Update indices and compile closures for agent delivery
            await self.conductor_setup(assigned=set(self.tps))

            # send first message
            await self.put(key=str(0).encode(), value=str(0).encode())

            # Wait for first message to be processed
            await self.wait(self.agent_started_processing, timeout=10.0)

            # send the rest of the messages
            for i in range(1, self.num_messages):
                await self.put(key=str(i).encode(), value=str(i).encode())

            # Wait for last message to be processed
            await self.wait(self.agent_stopped_processing, timeout=1000.0)

        self.finished.set()

    async def conductor_setup(self, assigned: Set[TP]) -> None:
        await self.app.agents.on_rebalance(set(), assigned)
        await self.app.topics._update_indices()
        await self.app.topics.on_partitions_assigned(assigned)

    async def put(self, key: bytes, value: bytes, *,
                  tp: TP = None,
                  **kwargs: Any) -> MessageT:
        # send first message
        message = self.Message(
            tp=tp, key=key, value=bytes, headers=None, **kwargs)
        await self.app.topics.on_message(message)
        return message

    def Message(self,
                tp: TP = None,
                offset: int = None,
                timestamp: float = None,
                timestamp_type: int = 1,
                headers: Optional[HeadersArg] = None,
                key: Optional[bytes] = None,
                value: Optional[bytes] = None,
                checksum: Optional[bytes] = None,
                **kwargs: Any) -> MessageT:
        if tp is None:
            tp = next(self.next_tp)
        return MessageT(
            topic=tp.topic,
            partition=tp.partition,
            offset=self.next_offset(tp) if offset is None else offset,
            timestamp=time() if timestamp is None else timestamp,
            timestamp_type=timestamp_type,
            headers=headers,
            key=key,
            value=value,
            checksum=checksum,
            tp=tp,
        )

    def next_offset(self, tp: TP, *, offsets=CURRENT_OFFSETS) -> int:
        offset = offsets[tp]
        offsets[tp] += 1
        return offset



"""Redis cache backend."""
import socket
import typing

from enum import Enum
from typing import Any, Mapping, Optional, Type, Union

from mode.utils.compat import want_bytes
from mode.utils.objects import cached_property
from yarl import URL

from faust.exceptions import ImproperlyConfigured
from faust.types import AppT
from . import base

try:
    import aredis
    import aredis.exceptions
except ImportError:  # pragma: no cover
    aredis = None  # noqa

if typing.TYPE_CHECKING:  # pragma: no cover
    from aredis import StrictRedis as _RedisClientT
else:
    class _RedisClientT: ...  # noqa


class RedisScheme(Enum):
    """Types of Redis configurations."""

    SINGLE_NODE = 'redis'
    CLUSTER = 'rediscluster'


class CacheBackend(base.CacheBackend):
    """Backend for cache operations using Redis."""

    connect_timeout: Optional[float]
    stream_timeout: Optional[float]
    max_connections: Optional[int]
    max_connections_per_node: Optional[int]

    _client: Optional[_RedisClientT] = None
    _client_by_scheme: Mapping[str, Type[_RedisClientT]]

    if aredis is None:  # pragma: no cover
        ...
    else:
        operational_errors = (
            socket.error,
            IOError,
            OSError,
            aredis.exceptions.ConnectionError,
            aredis.exceptions.TimeoutError,
        )
        invalidating_errors = (
            aredis.exceptions.DataError,
            aredis.exceptions.InvalidResponse,
            aredis.exceptions.ResponseError,
        )
        irrecoverable_errors = (
            aredis.exceptions.AuthenticationError,
        )

    def __init__(self,
                 app: AppT,
                 url: Union[URL, str],
                 *,
                 connect_timeout: float = None,
                 stream_timeout: float = None,
                 max_connections: int = None,
                 max_connections_per_node: int = None,
                 **kwargs: Any) -> None:
        super().__init__(app, url, **kwargs)
        self.connect_timeout = connect_timeout
        self.stream_timeout = stream_timeout
        self.max_connections = max_connections
        self.max_connections_per_node = max_connections_per_node
        self._client_by_scheme = self._init_schemes()

    def _init_schemes(self) -> Mapping[str, Type[_RedisClientT]]:
        if aredis is None:  # pragma: no cover
            return {}
        else:
            return {
                RedisScheme.SINGLE_NODE.value: aredis.StrictRedis,
                RedisScheme.CLUSTER.value: aredis.StrictRedisCluster,
            }

    async def _get(self, key: str) -> Optional[bytes]:
        value: Optional[bytes] = await self.client.get(key)
        if value is not None:
            return want_bytes(value)
        return None

    async def _set(self, key: str, value: bytes,
                   timeout: float = None) -> None:
        if timeout is not None:
            await self.client.setex(key, int(timeout), value)
        else:
            await self.client.set(key, value)

    async def _delete(self, key: str) -> None:
        await self.client.delete(key)

    async def on_start(self) -> None:
        """Call when Redis backend starts."""
        if aredis is None:
            raise ImproperlyConfigured(
                'Redis cache backend requires `pip install aredis`')
        await self.connect()

    async def connect(self) -> None:
        """Connect to Redis/Redis Cluster server."""
        if self._client is None:
            self._client = self._new_client()
        await self.client.ping()

    def _new_client(self) -> _RedisClientT:
        return self._client_from_url_and_query(self.url, **self.url.query)

    def _client_from_url_and_query(
            self,
            url: URL,
            *,
            connect_timeout: str = None,
            stream_timeout: str = None,
            max_connections: str = None,
            max_connections_per_node: str = None,
            **kwargs: Any) -> _RedisClientT:
        Client = self._client_by_scheme[url.scheme]
        return Client(**self._prepare_client_kwargs(
            url,
            host=url.host,
            port=url.port,
            db=self._db_from_path(url.path),
            password=url.password,
            connect_timeout=self._float_from_str(
                connect_timeout, self.connect_timeout),
            stream_timeout=self._float_from_str(
                stream_timeout, self.stream_timeout),
            max_connections=self._int_from_str(
                max_connections, self.max_connections),
            max_connections_per_node=self._int_from_str(
                max_connections_per_node, self.max_connections_per_node),
            skip_full_coverage_check=True,
        ))

    def _prepare_client_kwargs(self, url: URL, **kwargs: Any) -> Mapping:
        if url.scheme == RedisScheme.CLUSTER.value:
            return self._as_cluster_kwargs(**kwargs)
        return kwargs

    def _as_cluster_kwargs(self,
                           db: str = None,
                           **kwargs: Any) -> Mapping:
        # Redis Cluster does not support db as argument.
        return kwargs

    def _int_from_str(self,
                      val: str = None,
                      default: int = None) -> Optional[int]:
        return int(val) if val else default

    def _float_from_str(self,
                        val: str = None,
                        default: float = None) -> Optional[float]:
        return float(val) if val else default

    def _db_from_path(self, path: str) -> int:
        if not path or path == '/':
            return 0  # default db
        try:
            return int(path.strip('/'))
        except ValueError:
            raise ValueError(
                f'Database is int between 0 and limit - 1, not {path!r}')

    @cached_property
    def client(self) -> _RedisClientT:
        """Return Redis client instance."""
        if self._client is None:
            raise RuntimeError('Cache backend not started')
        return self._client



import asyncio
import io
import os
import sys
from contextlib import ExitStack, redirect_stderr, redirect_stdout
import pytest
from mode.utils.mocks import patch


def test_main(*, app, loop):
    from proj324 import main

    neu_loop = asyncio.set_event_loop(asyncio.new_event_loop())  # noqa

    # must use the event loop fixture to ensure not using old loop.
    stdout = io.StringIO()
    stderr = io.StringIO()
    environ = dict(os.environ)
    try:
        with ExitStack() as stack:
            stack.enter_context(patch(
                'sys.argv',
                ['proj', 'myprocesscommandi324']))
            assert sys.argv == ['proj', 'myprocesscommandi324']
            stack.enter_context(pytest.raises(SystemExit))
            stack.enter_context(redirect_stdout(stdout))
            stack.enter_context(redirect_stderr(stderr))

            main()
    finally:
        os.environ.clear()
        os.environ.update(environ)
    print(f'STDOUT: {stdout.getvalue()!r}')
    print(f'STDERR: {stderr.getvalue()!r}')
    assert 'IMPORTS __MAIN__' not in stdout.getvalue()
    assert 'TEST FILE IMPORTED' not in stdout.getvalue()
    assert 'HELLO WORLD #324' in stdout.getvalue()



from datetime import datetime
from decimal import Decimal
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)
import pytest
from faust.models import Record
from faust.models.typing import NodeType, TypeExpression
from faust.types import ModelT


class X(Record, namespace='test.X'):
    val: int


class Y(Record, namespace='test.Y'):
    val: str


class Z(NamedTuple):
    foo: X
    bar: int
    baz: str


class TypeExpressionTest(NamedTuple):
    type_expression: Type
    serialized_data: Any
    expected_result: Any
    expected_comprehension: str
    expected_types: Dict[NodeType, Set[Type]]


def Xi(i):
    return X(i).dumps()


CASE_TUPLE_LIST_SET_X = TypeExpressionTest(
    type_expression=Tuple[X, List[Set[X]]],
    serialized_data=[
        Xi(0),
        [[Xi(1), Xi(2), Xi(3)],
         [Xi(4), Xi(5), Xi(6)],
         [Xi(7), Xi(8)]]],
    expected_result=(
        X(0),
        [
            {X(1), X(2), X(3)},
            {X(4), X(5), X(6)},
            {X(7), X(8)},
        ],
    ),
    expected_comprehension='''\
(test__X._from_data_field(a[0]), \
[{test__X._from_data_field(c) for c in b} \
for b in a[1]])''',
    expected_types={NodeType.MODEL: {X}},
)

CASE_LIST_LIST_X = TypeExpressionTest(
    type_expression=List[List[X]],
    serialized_data=[
        [Xi(1), Xi(2), Xi(3)],
        [Xi(4), Xi(5), Xi(6)],
        [Xi(7), Xi(8), Xi(9)]],
    expected_result=[
        [X(1), X(2), X(3)],
        [X(4), X(5), X(6)],
        [X(7), X(8), X(9)],
    ],
    expected_comprehension='''\
[[test__X._from_data_field(c) for c in b] for b in a]''',
    expected_types={NodeType.MODEL: {X}},
)

CASE_DICT_KEY_STR_VALUE_OPTIONAL_SET_X_COMP = '''
    {b: ({test__X._from_data_field(d) for d in c} if c is not None else None) \
for b, c in a.items()}
'''.strip()
CASE_DICT_KEY_STR_VALUE_OPTIONAL_SET_X = TypeExpressionTest(
    type_expression=Dict[str, Optional[Set[X]]],
    serialized_data={
        'foo': [Xi(1), Xi(2), Xi(3)],
        'bar': [Xi(3), Xi(4), Xi(5)],
        'baz': [Xi(7), Xi(8)],
        'xaz': None,
        'xuz': [],
    },
    expected_result={
        'foo': {X(1), X(2), X(3)},
        'bar': {X(3), X(4), X(5)},
        'baz': {X(7), X(8)},
        'xaz': None,
        'xuz': set(),
    },
    expected_comprehension=CASE_DICT_KEY_STR_VALUE_OPTIONAL_SET_X_COMP,
    expected_types={NodeType.MODEL: {X}},
)

CASE_DICT_KEY_STR_VALUE_OPTIONAL_SET_ABSMODEL_COMP = '''
    {b: ({_Model_._from_data_field(d) for d in c} if c is not None else None) \
for b, c in a.items()}
'''.strip()
CASE_DICT_KEY_STR_VALUE_OPTIONAL_SET_ABSMODEL = TypeExpressionTest(
    type_expression=Dict[str, Optional[Set[ModelT]]],
    serialized_data={
        'foo': [Xi(1), Xi(2), Xi(3)],
        'bar': [Xi(3), Xi(4), Xi(5)],
        'baz': [Xi(7), Xi(8)],
        'xaz': None,
        'xuz': [],
    },
    expected_result={
        'foo': {X(1), X(2), X(3)},
        'bar': {X(3), X(4), X(5)},
        'baz': {X(7), X(8)},
        'xaz': None,
        'xuz': set(),
    },
    expected_comprehension=CASE_DICT_KEY_STR_VALUE_OPTIONAL_SET_ABSMODEL_COMP,
    expected_types={NodeType.MODEL: {ModelT}},
)
# List[Dict[Tuple[int, X], List[Mapping[str, Optional[Set[X]]]]]
CASE_COMPLEX1_COMP = '''\
[{(c[0], c[1]): [\
{f: ({test__X._from_data_field(h) for h in g} if g is not None else None) \
for f, g in e.items()} for e in d] for c, d in b.items()} for b in a]\
'''
CASE_COMPLEX1 = TypeExpressionTest(
    type_expression=List[Dict[
        Tuple[int, Any],
        List[Mapping[str, Optional[Set[X]]]],
    ]],
    serialized_data=[
        {
            (0, 1): [
                {'foo': [Xi(1), Xi(2), Xi(3), Xi(4)]},
                {'bar': [Xi(1), Xi(2)]},
                {'baz': None},
                {'xuz': []},
            ],
            (10, 20): [
                {'moo': [Xi(3), Xi(4), Xi(5), Xi(6)]},
            ],
        },
        {
            (30, 42131): [
                {'xfoo': [Xi(3), Xi(2), Xi(300), Xi(1012012)]},
                {'iasiqwoqwidfaoiwqh': [Xi(3120120), Xi(34894891892398)]},
                {'ieqieiai': None},
                {'uidafjaaoz': []},
            ],
            (3192, 12321): [
                {'moo': [Xi(3), Xi(4), Xi(5), Xi(6)]},
            ],
        },
    ],
    expected_result=[
        {
            (0, 1): [
                {'foo': {X(1), X(2), X(3), X(4)}},
                {'bar': {X(1), X(2)}},
                {'baz': None},
                {'xuz': set()},
            ],
            (10, 20): [
                {'moo': {X(3), X(4), X(5), X(6)}},
            ],
        },
        {
            (30, 42131): [
                {'xfoo': {X(3), X(2), X(300), X(1012012)}},
                {'iasiqwoqwidfaoiwqh': {X(3120120), X(34894891892398)}},
                {'ieqieiai': None},
                {'uidafjaaoz': set()},
            ],
            (3192, 12321): [
                {'moo': {X(3), X(4), X(5), X(6)}},
            ],
        },
    ],
    expected_comprehension=CASE_COMPLEX1_COMP,
    expected_types={NodeType.MODEL: {X}},
)


CASE_DICT_KEY_STR_VALUE_SET_NAMEDTUPLE_Z = TypeExpressionTest(
    type_expression=Dict[str, Set[Z]],
    serialized_data={
        'foo': [[Xi(1), 2, 'foo']],
        'bar': [[Xi(3), 4, 'bar1'], [Xi(5), 6, 'bar2'], [Xi(7), 8, 'bar3']],
        'baz': [],
    },
    expected_result={
        'foo': {Z(X(1), 2, 'foo')},
        'bar': {Z(X(3), 4, 'bar1'), Z(X(5), 6, 'bar2'), Z(X(7), 8, 'bar3')},
        'baz': set(),
    },
    expected_comprehension=None,
    expected_types={NodeType.MODEL: {X}},
)


CASE_SCALAR_STR = TypeExpressionTest(
    type_expression=str,
    serialized_data='foo',
    expected_result='foo',
    expected_comprehension='a',
    expected_types={},
)

CASE_OPTIONAL_SCALAR_STR = TypeExpressionTest(
    type_expression=str,
    serialized_data=None,
    expected_result=None,
    expected_comprehension='a',
    expected_types={},
)

CASE_SCALAR_INT = TypeExpressionTest(
    type_expression=int,
    serialized_data=100,
    expected_result=100,
    expected_comprehension='a',
    expected_types={},
)

CASE_LIST_INT = TypeExpressionTest(
    type_expression=List[int],
    serialized_data=[1, 2, 3, 4],
    expected_result=[1, 2, 3, 4],
    expected_comprehension='[b for b in a]',
    expected_types={},
)

CASE_LIST_DECIMAL = TypeExpressionTest(
    type_expression=List[Decimal],
    serialized_data=[
        str(Decimal('1.12')),
        str(Decimal('2.23')),
        str(Decimal('3.34')),
        str(Decimal('4.45')),
    ],
    expected_result=[
        Decimal('1.12'),
        Decimal('2.23'),
        Decimal('3.34'),
        Decimal('4.45'),
    ],
    expected_comprehension='[_Decimal_(b) for b in a]',
    expected_types={NodeType.DECIMAL: {Decimal}},
)

CASE_MAPPING_KEY_DATETIME_VALUE_OPTIONAL_LIST_DATETIME = TypeExpressionTest(
    type_expression=Mapping[datetime, Optional[List[datetime]]],
    serialized_data={
        datetime(2019, 12, 16, 19, 45, 18, 162288).isoformat(): [
            datetime(2017, 12, 16, 19, 45, 18, 162288).isoformat(),
            datetime(2018, 12, 16, 19, 45, 18, 162288).isoformat(),
        ],
        datetime(2019, 12, 15, 11, 45, 18, 9391).isoformat(): None,
        datetime(2018, 12, 15, 11, 45, 18, 9391).isoformat(): [],
    },
    expected_result={
        datetime(2019, 12, 16, 19, 45, 18, 162288): [
            datetime(2017, 12, 16, 19, 45, 18, 162288),
            datetime(2018, 12, 16, 19, 45, 18, 162288),
        ],
        datetime(2019, 12, 15, 11, 45, 18, 9391): None,
        datetime(2018, 12, 15, 11, 45, 18, 9391): [],
    },
    expected_comprehension='''\
{_iso8601_parse_(b): \
([_iso8601_parse_(d) for d in c] if c is not None else None) \
for b, c in a.items()}''',
    expected_types={NodeType.DATETIME: {datetime}},
)

CASE_LIST_OPTIONAL_INT = TypeExpressionTest(
    type_expression=List[Optional[int]],
    serialized_data=[1, 2, 3, None, 4],
    expected_result=[1, 2, 3, None, 4],
    expected_comprehension='[(b if b is not None else None) for b in a]',
    expected_types={},
)

CASE_SCALAR_MODEL = TypeExpressionTest(
    type_expression=X,
    serialized_data=X(10).dumps(),
    expected_result=X(10),
    expected_comprehension='test__X._from_data_field(a)',
    expected_types={NodeType.MODEL: {X}},
)

CASE_SCALAR_OPTIONAL_MODEL = TypeExpressionTest(
    type_expression=Optional[X],
    serialized_data=X(10).dumps(),
    expected_result=X(10),
    expected_comprehension='''\
(test__X._from_data_field(a) if a is not None else None)''',
    expected_types={NodeType.MODEL: {X}},
)

CASE_SCALAR_ABSTRACT_MODEL = TypeExpressionTest(
    type_expression=ModelT,
    serialized_data=X(10).dumps(),
    expected_result=X(10),
    expected_comprehension='_Model_._from_data_field(a)',
    expected_types={NodeType.MODEL: {ModelT}},
)

CASE_SCALAR_OPTIONAL_ABSTRACT_MODEL = TypeExpressionTest(
    type_expression=Union[ModelT, None],
    serialized_data=X(10).dumps(),
    expected_result=X(10),
    expected_comprehension='''\
(_Model_._from_data_field(a) if a is not None else None)''',
    expected_types={NodeType.MODEL: {ModelT}},
)

CASE_SCALAR_DECIMAL = TypeExpressionTest(
    type_expression=Decimal,
    serialized_data=str(Decimal('120213.123123')),
    expected_result=Decimal('120213.123123'),
    expected_comprehension='_Decimal_(a)',
    expected_types={NodeType.DECIMAL: {Decimal}},
)

NOW = datetime.utcnow()
CASE_SCALAR_DATETIME = TypeExpressionTest(
    type_expression=datetime,
    serialized_data=NOW.isoformat(),
    expected_result=NOW,
    expected_comprehension='_iso8601_parse_(a)',
    expected_types={NodeType.DATETIME: {datetime}},
)

CASE_TUPLE_SINGLE_ELEMENT = TypeExpressionTest(
    type_expression=Tuple[int],
    serialized_data=[1, 2, 3],
    expected_result=(1,),
    expected_comprehension='(a[0],)',
    expected_types={},
)


CASE_TUPLE_X_VARARGS = TypeExpressionTest(
    type_expression=Tuple[X, ...],
    serialized_data=[Xi(1), Xi(2), Xi(3)],
    expected_result=(X(1), X(2), X(3)),
    expected_comprehension='tuple(test__X._from_data_field(b) for b in a)',
    expected_types={NodeType.MODEL: {X}},
)

CASE_TUPLE_NO_ARGS = TypeExpressionTest(
    type_expression=tuple,
    serialized_data=[1, 2, 3],
    expected_result=(1, 2, 3),
    expected_comprehension='tuple(a)',
    expected_types={},
)

CASE_LIST_NO_ARGS = TypeExpressionTest(
    type_expression=list,
    serialized_data=[1, 2, 3],
    expected_result=[1, 2, 3],
    expected_comprehension='list(a)',
    expected_types={},
)

CASE_UNION_STR_INT_FLOAT = TypeExpressionTest(
    type_expression=Union[str, int, float],
    serialized_data='foo',
    expected_result='foo',
    expected_comprehension='a',
    expected_types={},
)

CASE_UNION_X_Y = TypeExpressionTest(
    type_expression=Union[X, Y],
    serialized_data=Y('foo').dumps(),
    expected_result=Y('foo'),
    expected_comprehension='_Model_._from_data_field(a)',
    expected_types={NodeType.MODEL: {ModelT}},
)

CASES = [
    CASE_TUPLE_LIST_SET_X,
    CASE_LIST_LIST_X,
    CASE_DICT_KEY_STR_VALUE_OPTIONAL_SET_X,
    CASE_DICT_KEY_STR_VALUE_OPTIONAL_SET_ABSMODEL,
    CASE_COMPLEX1,
    CASE_DICT_KEY_STR_VALUE_SET_NAMEDTUPLE_Z,
    CASE_SCALAR_STR,
    CASE_OPTIONAL_SCALAR_STR,
    CASE_SCALAR_INT,
    CASE_LIST_INT,
    CASE_LIST_DECIMAL,
    CASE_MAPPING_KEY_DATETIME_VALUE_OPTIONAL_LIST_DATETIME,
    CASE_LIST_OPTIONAL_INT,
    CASE_SCALAR_MODEL,
    CASE_SCALAR_OPTIONAL_MODEL,
    CASE_SCALAR_ABSTRACT_MODEL,
    CASE_SCALAR_OPTIONAL_ABSTRACT_MODEL,
    CASE_SCALAR_DECIMAL,
    CASE_SCALAR_DATETIME,
    CASE_TUPLE_SINGLE_ELEMENT,
    CASE_TUPLE_X_VARARGS,
    CASE_TUPLE_NO_ARGS,
    CASE_LIST_NO_ARGS,
    CASE_UNION_STR_INT_FLOAT,
    CASE_UNION_X_Y,
]


@pytest.mark.parametrize('case', CASES)
def test_comprehension(case):
    expr = TypeExpression(case.type_expression)
    if case.expected_comprehension:
        assert expr.as_comprehension() == case.expected_comprehension


@pytest.mark.parametrize('case', CASES)
def test_compile(case):
    expr = TypeExpression(case.type_expression)
    fun = expr.as_function(globals=globals())
    assert fun(case.serialized_data) == case.expected_result
    assert expr.found_types == case.expected_types



#!/usr/bin/env python
import random
import signal
from itertools import cycle
from typing import List, NamedTuple

import envoy
from mode import Service, Worker


class Span(NamedTuple):
    length: int
    signals: List[int]
    min_sleep: float
    max_sleep: float

    def seconds_to_sleep(self):
        return random.uniform(self.min_sleep, self.max_sleep)


class Chaos(Service):

    schedule = [
        # Signal -TERM/-INT between every 1 and 30 seconds.
        # This period lasts for at least half a minute, but never for more
        # than 50 minutes.
        Span(
            100,  # hundred times sleep between random(1, 30) seconds.
            signals=[signal.SIGTERM, signal.SIGINT],
            min_sleep=1.0,
            max_sleep=30.0,
        ),
        # Signal -TERM between every 5 and 100 seconds.
        # lasts for at least 4.1 minutes, at most 83 minutes.
        Span(
            50,  # fifty times sleep between random(5, 100) seconds.
            signals=[signal.SIGTERM],
            min_sleep=5.0,
            max_sleep=100.0,
        ),
        # super fast burst of signal -9 (between every 0.1s and 1.0 second).
        # lasts for at most 30 seconds.
        # This emulates what happens in production sometimes when the OOM
        # killer is activated.
        Span(
            30,  # for thirty times sleep between random(0.1, 1.0) seconds.
            signals=[signal.SIGKILL],
            min_sleep=0.1,
            max_sleep=1.0,
        ),
        # we repeat here forever, (see iterate_over_scheduled_time()) below.)
        # --
    ]

    def iterate_over_scheduled_time(self):
        for period in cycle(self.schedule):
            for _ in range(period.length):
                yield period

    @Service.task
    async def _runs_in_background(self):
        self.log.info("I'm your local friendly chaos monkey tester")
        self.log.info('Starting signal dispatcher.')
        for current_span in self.iterate_over_scheduled_time():
            if self.should_stop:
                return
            secs = current_span.seconds_to_sleep()
            self.log.info('Signal dispatcher sleeping for %r seconds...', secs)
            await self.sleep(secs)
            sig = random.choice(current_span.signals)
            self.log.warning('Signalling all workers on this box with %r', sig)
            r = envoy.run(f'pkill -{int(sig)} Faust:Worker')
            if r.status_code:
                if r.std_err.strip():
                    self.log.error('ERROR from pkill: %r', r.std_err)
                else:
                    self.log.info('No processes running, nothing to signal!')


if __name__ == '__main__':
    Worker(Chaos(), loglevel='INFO').execute_from_commandline()



#!/usr/bin/env python

# In this exapmple we have a function `publish_every_2secs` publishing a
# message every 2 seconds to topic `tumbling_topic`
# We have created an agent `print_windowed_events` consuming events from
# `tumbling_topic` that mutates the windowed table `tumbling_table`

# `tumbling_table` is a table with tumbling (non overlaping) windows. Each of
# its windows is 10 seconds of duration. The table counts the number of events
# per window. Since we produce an event every 2 seconds and our windows are 10
# seconds of duration we expect different the following results per method
# called in `WindowWrapper`:
# - now(): Gets the closest value to current local time. From 1 to 5
# - current(): Gets the value relative to the event's timestamp. From 1 to 5.
# - value(): Gets the value relative to default relative option
# - delta(30): Gets the value of window 30 secs before the current event. For
# the first 30 seconds it will be 0 and after second 40 it will always be 5.

from random import random
from datetime import timedelta
import faust

app = faust.App('windowing', broker='kafka://localhost:9092')


class Model(faust.Record, serializer='json'):
    random: float


TOPIC = 'tumbling_topic'

tumbling_topic = app.topic(TOPIC, value_type=Model)
tumbling_table = app.Table(
    'tumbling_table',
    default=int).tumbling(10, expires=timedelta(minutes=10))


@app.agent(tumbling_topic)
async def print_windowed_events(stream):
    async for _ in stream: # noqa
        tumbling_table['counter'] += 1

        value = tumbling_table['counter']

        print('-- New Event (every 2 secs) written to tumbling(10) --')
        print(f'now() should go from 1 to 5: {value.now()}')
        print(f'current() should go from 1 to 5: {value.current()}')
        print(f'value() should go from 1 to 5: {value.value()}')
        print('delta(30) should start at 0 and after 40 secs be 5: '
              f'{value.delta(30)}')


@app.timer(2.0, on_leader=True)
async def publish_every_2secs():
    msg = Model(random=round(random(), 2))
    await tumbling_topic.send(value=msg)


if __name__ == '__main__':
    app.main()



"""HTTP endpoint showing statistics from the Faust monitor."""
from typing import List, MutableMapping
from faust import web

__all__ = ['Index', 'blueprint']

TPMap = MutableMapping[str, List[int]]


blueprint = web.Blueprint('production_index')


@blueprint.route('/', name='index')
class Index(web.View):
    """Dummy Index."""

    async def get(self, request: web.Request) -> web.Response:
        """Simple status page."""
        return self.json({'status': 'OK'})



import pytest
from faust.utils.codegen import reprcall, reprkwargs


@pytest.mark.parametrize('input,expected', [
    ({}, ''),
    ({'foo': 'bar'}, "foo='bar'"),
    ({'foo': 'bar', 'xaz': 300.3}, "foo='bar', xaz=300.3"),
])
def test_reprkwargs(input, expected):
    assert reprkwargs(input) == expected


@pytest.mark.parametrize('name,args,kwargs,expected', [
    ('NAME', (), {}, 'NAME()'),
    ('NAME', (1, 2.3, 3), {}, 'NAME(1, 2.3, 3)'),
    ('NAME', (1, 2.3, 3), {'foo': 'bar'}, "NAME(1, 2.3, 3, foo='bar')"),
    ('NAME', (1, 2.3, 3), {'foo': 'bar', 'xaz': 300.3},
     "NAME(1, 2.3, 3, foo='bar', xaz=300.3)"),
])
def test_reprcall(name, args, kwargs, expected):
    assert reprcall(name, args, kwargs) == expected



import socket
import sys
from pathlib import Path

import faust
import mode
import pytest
import pytz

from mode.utils.mocks import patch
from yarl import URL

from faust import App
from faust.app import BootStrategy
from faust.assignor import LeaderAssignor, PartitionAssignor
from faust.exceptions import AlreadyConfiguredWarning, ImproperlyConfigured
from faust.app.router import Router
from faust.sensors import Monitor
from faust.serializers import Registry
from faust.tables import TableManager
from faust.transport.utils import DefaultSchedulingStrategy
from faust.types import settings
from faust.types.enums import ProcessingGuarantee
from faust.types.web import ResourceOptions

TABLEDIR: Path
DATADIR: Path
if sys.platform == 'win32':
    TABLEDIR = Path('c:/Program Files/Faust/')
    DATADIR = Path('c:/Temporary Files/Faust/')
else:
    DATADIR = Path('/etc/faust/')
    TABLEDIR = Path('/var/faust/')


class OtherSchedulingStrategy(DefaultSchedulingStrategy):
    ...


def _dummy_partitioner(a, b, c):
    return 0


class test_settings:

    def App(self, id='myid', **kwargs):
        app = App(id, **kwargs)
        app.finalize()
        return app

    def test_defaults(self):
        app = self.App()
        conf = app.conf
        assert not conf.debug
        assert conf.broker == [URL(settings.BROKER_URL)]
        assert conf.broker_consumer == [URL(settings.BROKER_URL)]
        assert conf.broker_producer == [URL(settings.BROKER_URL)]
        assert conf.store == URL(settings.STORE_URL)
        assert conf.cache == URL(settings.CACHE_URL)
        assert conf.web == URL(settings.WEB_URL)
        assert conf.web_enabled
        assert not conf.web_in_thread
        assert conf.datadir == conf._prepare_datadir(settings.DATADIR)
        assert conf.tabledir == conf._prepare_tabledir(settings.TABLEDIR)
        assert conf.processing_guarantee == settings.PROCESSING_GUARANTEE
        assert conf.broker_api_version == settings.BROKER_API_VERSION
        assert conf.broker_client_id == settings.BROKER_CLIENT_ID
        assert conf.broker_request_timeout == settings.BROKER_REQUEST_TIMEOUT
        assert conf.broker_session_timeout == settings.BROKER_SESSION_TIMEOUT
        assert (conf.broker_rebalance_timeout ==
                settings.BROKER_REBALANCE_TIMEOUT)
        assert (conf.broker_heartbeat_interval ==
                settings.BROKER_HEARTBEAT_INTERVAL)
        assert conf.broker_commit_interval == settings.BROKER_COMMIT_INTERVAL
        assert conf.broker_commit_every == settings.BROKER_COMMIT_EVERY
        assert (conf.broker_commit_livelock_soft_timeout ==
                settings.BROKER_LIVELOCK_SOFT)
        assert conf.broker_check_crcs
        assert conf.consumer_api_version == settings.BROKER_API_VERSION
        assert conf.timezone is settings.TIMEZONE
        assert conf.table_cleanup_interval == settings.TABLE_CLEANUP_INTERVAL
        assert conf.table_key_index_size == settings.TABLE_KEY_INDEX_SIZE
        assert conf.reply_to_prefix == settings.REPLY_TO_PREFIX
        assert conf.reply_expires == settings.REPLY_EXPIRES
        assert conf.stream_buffer_maxsize == settings.STREAM_BUFFER_MAXSIZE
        assert conf.stream_recovery_delay == settings.STREAM_RECOVERY_DELAY
        assert conf.stream_processing_timeout == (
            settings.STREAM_PROCESSING_TIMEOUT)
        assert conf.producer_partitioner is None
        assert (conf.producer_request_timeout ==
                settings.PRODUCER_REQUEST_TIMEOUT)
        assert conf.producer_api_version == settings.BROKER_API_VERSION
        assert (conf.stream_publish_on_commit ==
                settings.STREAM_PUBLISH_ON_COMMIT)
        assert conf.stream_wait_empty
        assert (conf.broker_max_poll_records ==
                settings.BROKER_MAX_POLL_RECORDS)
        assert (conf.broker_max_poll_interval ==
                settings.BROKER_MAX_POLL_INTERVAL)
        assert (conf.consumer_auto_offset_reset ==
                settings.CONSUMER_AUTO_OFFSET_RESET)
        assert not conf.autodiscover
        assert conf.origin is None
        assert conf.key_serializer == 'raw'
        assert conf.value_serializer == 'json'
        assert conf.reply_to is not None
        assert not conf.reply_create_topic
        assert conf.table_standby_replicas == 1
        assert conf.topic_replication_factor == 1
        assert conf.topic_partitions == 8
        assert conf.topic_allow_declare
        assert not conf.topic_disable_leader
        assert conf.logging_config is None
        assert conf.loghandlers == []
        assert conf.version == 1
        assert conf.canonical_url == URL(f'http://{socket.gethostname()}:6066')
        assert conf.web_bind == '0.0.0.0'
        assert conf.web_port == 6066
        assert conf.web_transport == settings.WEB_TRANSPORT
        assert conf.web_cors_options is None
        assert conf.worker_redirect_stdouts
        assert conf.worker_redirect_stdouts_level == 'WARN'

        assert conf.agent_supervisor is mode.OneForOneSupervisor

        assert conf.Agent is faust.Agent
        assert conf.ConsumerScheduler is DefaultSchedulingStrategy
        assert conf.Event is faust.Event
        assert conf.Schema is faust.Schema
        assert conf.Stream is faust.Stream
        assert conf.Table is faust.Table
        assert conf.TableManager is TableManager
        assert conf.Serializers is Registry
        assert conf.Worker is faust.Worker
        assert conf.PartitionAssignor is PartitionAssignor
        assert conf.LeaderAssignor is LeaderAssignor
        assert conf.Router is Router
        assert conf.Topic is faust.Topic
        from aiohttp.client import ClientSession
        assert conf.HttpClient is ClientSession
        assert conf.Monitor is Monitor

    def test_reply_prefix_unique(self):
        app1 = self.App()
        app2 = self.App()
        assert app1.conf.reply_to != app2.conf.reply_to

    def test_app_config(self):
        self.assert_config_equivalent()

    def test_broker_as_URL(self):
        app = self.assert_config_equivalent(broker='ckafka://')
        assert isinstance(app.conf.broker, list)
        assert app.conf.broker[0] == URL('ckafka://')

    def test_store_as_URL(self):
        app = self.assert_config_equivalent(store=URL('moo://'))
        assert isinstance(app.conf.store, URL)

    def test_cache_as_URL(self):
        app = self.assert_config_equivalent(cache=URL('moo://'))
        assert isinstance(app.conf.cache, URL)

    def test_web_as_URL(self):
        app = self.assert_config_equivalent(web=URL('moo://'))
        assert isinstance(app.conf.web, URL)

    def test_datadir_as_Path(self):
        app = self.assert_config_equivalent(datadir=DATADIR)
        assert isinstance(app.conf.datadir, Path)

    def test_tabledir_is_relative_to_path(self):
        app = self.assert_config_equivalent(
            datadir=str(DATADIR),
            tabledir='moo',
        )
        assert app.conf.tabledir == app.conf.appdir / Path('moo')

    def assert_config_equivalent(self,
                                 id='id',
                                 version=303,
                                 broker='foo://',
                                 store='bar://',
                                 cache='baz://',
                                 web='xuzzy://',
                                 web_enabled=False,
                                 autodiscover=True,
                                 origin='faust',
                                 canonical_url='http://example.com/',
                                 broker_client_id='client id',
                                 datadir=str(DATADIR),
                                 tabledir=str(TABLEDIR),
                                 processing_guarantee='exactly_once',
                                 broker_api_version='0.1',
                                 broker_request_timeout=10000.05,
                                 broker_heartbeat_interval=101.13,
                                 broker_session_timeout=30303.30,
                                 broker_rebalance_timeout=606606.60,
                                 broker_commit_every=202,
                                 broker_commit_interval=30.3,
                                 broker_commit_livelock_soft_timeout=60.6,
                                 broker_check_crcs=False,
                                 broker_producer='moo://',
                                 broker_consumer='zoo://',
                                 consumer_api_version='0.4',
                                 producer_partitioner=_dummy_partitioner,
                                 producer_request_timeout=2.66,
                                 producer_api_version='0.10',
                                 table_cleanup_interval=80.8,
                                 table_key_index_size=1999,
                                 key_serializer='str',
                                 value_serializer='str',
                                 table_standby_replicas=48,
                                 topic_replication_factor=16,
                                 topic_allow_declare=False,
                                 topic_disable_leader=True,
                                 reply_to='reply_to',
                                 reply_create_topic=True,
                                 reply_expires=90.9,
                                 stream_buffer_maxsize=101,
                                 stream_wait_empty=True,
                                 stream_publish_on_commit=False,
                                 stream_recovery_delay=69.3,
                                 stream_processing_timeout=73.03,
                                 web_bind='localhost',
                                 web_port=6069,
                                 web_host='localhost',
                                 web_transport='udp://',
                                 web_in_thread=True,
                                 web_cors_options={  # noqa: B006
                                    'http://example.com': ResourceOptions(
                                        allow_credentials=True,
                                        expose_headers='*',
                                        allow_headers='*',
                                        max_age=3132,
                                        allow_methods='*',
                                    ),
                                 },
                                 worker_redirect_stdouts=False,
                                 worker_redirect_stdouts_level='DEBUG',
                                 broker_max_poll_records=1000,
                                 broker_max_poll_interval=10000,
                                 timezone=pytz.timezone('US/Eastern'),
                                 logging_config={'foo': 10},  # noqa
                                 consumer_auto_offset_reset='latest',
                                 ConsumerScheduler=OtherSchedulingStrategy,
                                 **kwargs) -> App:
        livelock_soft_timeout = broker_commit_livelock_soft_timeout
        app = self.App(
            id,
            version=version,
            broker=broker,
            broker_consumer=broker_consumer,
            broker_producer=broker_producer,
            store=store,
            cache=cache,
            web=web,
            web_enabled=web_enabled,
            autodiscover=autodiscover,
            origin=origin,
            canonical_url=canonical_url,
            broker_client_id=broker_client_id,
            datadir=datadir,
            tabledir=tabledir,
            processing_guarantee=processing_guarantee,
            broker_api_version=broker_api_version,
            broker_request_timeout=broker_request_timeout,
            broker_session_timeout=broker_session_timeout,
            broker_rebalance_timeout=broker_rebalance_timeout,
            broker_heartbeat_interval=broker_heartbeat_interval,
            broker_commit_every=broker_commit_every,
            broker_commit_interval=broker_commit_interval,
            broker_commit_livelock_soft_timeout=livelock_soft_timeout,
            broker_check_crcs=broker_check_crcs,
            broker_max_poll_records=broker_max_poll_records,
            broker_max_poll_interval=broker_max_poll_interval,
            consumer_api_version=consumer_api_version,
            producer_partitioner=producer_partitioner,
            producer_request_timeout=producer_request_timeout,
            producer_api_version=producer_api_version,
            table_cleanup_interval=table_cleanup_interval,
            table_key_index_size=table_key_index_size,
            key_serializer=key_serializer,
            value_serializer=value_serializer,
            table_standby_replicas=table_standby_replicas,
            topic_replication_factor=topic_replication_factor,
            topic_allow_declare=topic_allow_declare,
            topic_disable_leader=topic_disable_leader,
            reply_to=reply_to,
            reply_create_topic=reply_create_topic,
            reply_expires=reply_expires,
            stream_buffer_maxsize=stream_buffer_maxsize,
            stream_wait_empty=stream_wait_empty,
            stream_publish_on_commit=stream_publish_on_commit,
            stream_recovery_delay=stream_recovery_delay,
            stream_processing_timeout=stream_processing_timeout,
            timezone=timezone,
            web_bind=web_bind,
            web_port=web_port,
            web_host=web_host,
            web_transport=web_transport,
            web_in_thread=web_in_thread,
            web_cors_options=web_cors_options,
            worker_redirect_stdouts=worker_redirect_stdouts,
            worker_redirect_stdouts_level=worker_redirect_stdouts_level,
            logging_config=logging_config,
            consumer_auto_offset_reset=consumer_auto_offset_reset,
            ConsumerScheduler=ConsumerScheduler,
        )
        conf = app.conf
        assert conf.id == app.conf._prepare_id(id)
        assert conf.broker == [URL(broker)]
        assert conf.broker_consumer == [URL(broker_consumer)]
        assert conf.broker_producer == [URL(broker_producer)]
        assert conf.store == URL(str(store))
        assert conf.cache == URL(str(cache))
        assert conf.web == URL(str(web))
        assert not conf.web_enabled
        assert conf.web_in_thread
        assert conf.autodiscover == autodiscover
        assert conf.canonical_url == URL(str(canonical_url))
        assert conf.broker_client_id == broker_client_id
        assert conf.datadir == Path(str(datadir))
        if Path(tabledir).is_absolute():
            assert conf.tabledir == Path(str(tabledir))
        else:
            assert conf.tabledir.relative_to(conf.appdir) == Path(tabledir)
        assert conf.processing_guarantee == ProcessingGuarantee.EXACTLY_ONCE
        assert conf.broker_api_version == broker_api_version
        assert conf.broker_request_timeout == broker_request_timeout
        assert conf.broker_heartbeat_interval == broker_heartbeat_interval
        assert conf.broker_session_timeout == broker_session_timeout
        assert conf.broker_rebalance_timeout == broker_rebalance_timeout
        assert conf.broker_commit_every == broker_commit_every
        assert conf.broker_commit_interval == broker_commit_interval
        assert (conf.broker_commit_livelock_soft_timeout ==
                broker_commit_livelock_soft_timeout)
        assert conf.broker_check_crcs == broker_check_crcs
        assert conf.consumer_api_version == consumer_api_version
        assert conf.producer_partitioner is producer_partitioner
        assert conf.producer_request_timeout == producer_request_timeout
        assert conf.producer_api_version == producer_api_version
        assert conf.table_cleanup_interval == table_cleanup_interval
        assert conf.table_key_index_size == table_key_index_size
        assert conf.key_serializer == key_serializer
        assert conf.value_serializer == value_serializer
        assert conf.table_standby_replicas == table_standby_replicas
        assert conf.topic_replication_factor == topic_replication_factor
        assert conf.topic_allow_declare == topic_allow_declare
        assert conf.topic_disable_leader == topic_disable_leader
        assert conf.reply_to == reply_to
        assert conf.reply_expires == reply_expires
        assert conf.stream_buffer_maxsize == stream_buffer_maxsize
        assert conf.stream_wait_empty == stream_wait_empty
        assert conf.stream_publish_on_commit == stream_publish_on_commit
        assert conf.stream_recovery_delay == stream_recovery_delay
        assert conf.stream_processing_timeout == stream_processing_timeout
        assert conf.timezone is timezone
        assert conf.web_bind == web_bind
        assert conf.web_port == web_port
        assert conf.web_host == web_host
        assert conf.web_transport == URL(web_transport)
        assert conf.web_cors_options == web_cors_options
        assert conf.worker_redirect_stdouts == worker_redirect_stdouts
        assert (conf.worker_redirect_stdouts_level ==
                worker_redirect_stdouts_level)
        assert conf.broker_max_poll_records == broker_max_poll_records
        assert conf.broker_max_poll_interval == broker_max_poll_interval
        assert conf.logging_config == logging_config
        assert conf.consumer_auto_offset_reset == consumer_auto_offset_reset
        assert conf.ConsumerScheduler is OtherSchedulingStrategy
        return app

    def test_custom_host_port_to_canonical(self,
                                           web_bind='localhost',
                                           web_port=6069,
                                           web_host='localhost'):
        app = self.App(
            'id',
            web_bind=web_bind,
            web_port=web_port,
            web_host=web_host,
        )
        assert app.conf.canonical_url == URL(
            f'http://{app.conf.web_host}:{app.conf.web_port}')

    def test_id_no_version(self):
        assert self.App('id', version=1).conf.id == 'id'

    def test_version_cannot_be_zero(self):
        app = App('id', version=0)
        with pytest.raises(ImproperlyConfigured):
            app.finalize()

    def test_compat_url(self):
        assert self.App(url='foo').conf.broker == [URL('kafka://foo')]

    def test_compat_client_id(self):
        with pytest.warns(FutureWarning):
            assert self.App(client_id='foo').conf.broker_client_id == 'foo'

    def test_compat_commit_interval(self):
        with pytest.warns(FutureWarning):
            assert self.App(
                commit_interval=313.3).conf.broker_commit_interval == 313.3

    def test_compat_create_reply_topic(self):
        with pytest.warns(FutureWarning):
            assert self.App(create_reply_topic=True).conf.reply_create_topic
        with pytest.warns(FutureWarning):
            assert not self.App(
                create_reply_topic=False).conf.reply_create_topic

    def test_compat_num_standby_replicas(self):
        with pytest.warns(FutureWarning):
            assert self.App(
                num_standby_replicas=34).conf.table_standby_replicas == 34

    def test_compat_stream_ack_cancelled_tasks(self):
        with pytest.warns(UserWarning):
            assert not self.App(
                stream_ack_cancelled_tasks=False,
            ).conf.stream_ack_cancelled_tasks

    def test_compat_stream_ack_exceptions(self):
        with pytest.warns(UserWarning):
            assert self.App(
                stream_ack_exceptions=True).conf.stream_ack_exceptions

    def test_compat_default_partitions(self):
        with pytest.warns(FutureWarning):
            assert self.App(
                default_partitions=35).conf.topic_partitions == 35

    def test_compat_replication_factor(self):
        with pytest.warns(FutureWarning):
            assert self.App(
                replication_factor=36).conf.topic_replication_factor == 36

    def test_warns_when_key_already_configured(self):
        app = self.App(topic_partitions=37, topic_replication_factor=38)
        assert app.conf.topic_partitions == 37
        with pytest.warns(AlreadyConfiguredWarning):
            app.conf.topic_partitions = 39
        assert app.conf.topic_partitions == 39
        app.conf.topic_replication_factor = 40
        assert app.conf.topic_replication_factor == 40

    def test_broker_with_no_scheme_set(self):
        app = self.App(broker='example.com:3123')
        url = app.conf.broker[0]
        assert url.scheme == settings.DEFAULT_BROKER_SCHEME
        assert url.host == 'example.com'
        assert url.port == 3123

    def test_consumer_api_version__defaults_to_broker(self):
        expected_broker_version = '0.3333'
        app = self.App(
            broker_api_version=expected_broker_version,
            consumer_api_version=None,
        )
        assert app.conf.consumer_api_version == expected_broker_version

    def test_producer_api_version__defaults_to_broker(self):
        expected_broker_version = '0.3333'
        app = self.App(
            broker_api_version=expected_broker_version,
            producer_api_version=None,
        )
        assert app.conf.producer_api_version == expected_broker_version


class test_BootStrategy:

    def test_init(self, *, app):
        assert not BootStrategy(app, enable_web=False).enable_web
        assert BootStrategy(app, enable_web=True).enable_web
        assert not BootStrategy(app, enable_kafka=False).enable_kafka
        assert BootStrategy(app, enable_kafka=True).enable_kafka
        assert not BootStrategy(
            app, enable_kafka_producer=False,
        ).enable_kafka_producer
        assert BootStrategy(
            app, enable_kafka_producer=True,
        ).enable_kafka_producer
        assert not BootStrategy(
            app, enable_kafka_consumer=False,
        ).enable_kafka_consumer
        assert BootStrategy(
            app, enable_kafka_consumer=True,
        ).enable_kafka_consumer
        assert not BootStrategy(app, enable_sensors=False).enable_sensors
        assert BootStrategy(app, enable_sensors=True).enable_sensors

    def test_sensors(self, *, app):
        assert BootStrategy(app, enable_sensors=True).sensors() is app.sensors
        assert not BootStrategy(app, enable_sensors=False).sensors()

    def test_kafka_consumer(self, *, app):
        assert BootStrategy(
            app, enable_kafka_consumer=True).kafka_consumer()
        assert BootStrategy(
            app, enable_kafka_consumer=True).kafka_conductor()
        assert not BootStrategy(
            app, enable_kafka_consumer=False).kafka_consumer()
        assert not BootStrategy(
            app, enable_kafka_consumer=False).kafka_conductor()
        assert BootStrategy(
            app, enable_kafka=True).kafka_consumer()
        assert BootStrategy(
            app, enable_kafka=True).kafka_conductor()
        assert not BootStrategy(
            app, enable_kafka=False).kafka_consumer()
        assert not BootStrategy(
            app, enable_kafka=False).kafka_conductor()

    def test_kafka_producer(self, *, app):
        assert BootStrategy(
            app, enable_kafka_producer=True).kafka_producer()
        assert not BootStrategy(
            app, enable_kafka_producer=False).kafka_producer()
        assert BootStrategy(
            app, enable_kafka=True).kafka_producer()
        assert BootStrategy(
            app, enable_kafka_producer=None).kafka_producer()
        assert not BootStrategy(
            app, enable_kafka=False).kafka_producer()

    def test_web_server(self, *, app):
        assert BootStrategy(app, enable_web=True).web_server()
        assert not BootStrategy(app, enable_web=False).web_server()
        assert BootStrategy(app, enable_web=True).web_server()
        assert not BootStrategy(app, enable_web=False).web_server()
        assert BootStrategy(app, enable_web=None).web_server()

    def test_disable_kafka(self, *, app):
        class B(BootStrategy):
            enable_kafka = False

        b = B(app)
        assert not b.enable_kafka
        assert not b.kafka_conductor()
        assert not b.kafka_consumer()
        assert not b.kafka_producer()

    def test_disable_kafka_consumer(self, *, app):
        class B(BootStrategy):
            enable_kafka_consumer = False

        b = B(app)
        assert b.enable_kafka
        assert not b.kafka_conductor()
        assert not b.kafka_consumer()
        assert b.kafka_producer()

    @pytest.mark.app(debug=True)
    @pytest.mark.asyncio
    async def test_debug_enabled_warns_on_start(self, *, app):
        with patch('faust.app.base.logger') as logger:
            await app.on_start()
            logger.warning.assert_called_once()
